/*_############################################################################
  _## 
  _##  SNMP4J-Model - AbstractVariableBindingMapper.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/

package org.snmp4j.model.mapper;

import org.snmp4j.mp.SnmpConstants;
import org.snmp4j.smi.OctetString;
import org.snmp4j.smi.Variable;

import java.util.*;

/**
 * <code>AbstractVariableBindingMapper</code> provides helper functions to support the
 * conversion between {@link org.snmp4j.smi.Variable} values and native values used in an application.
 *
 * @author Frank Fock
 * @version 1.0
 */
public abstract class AbstractVariableBindingMapper<L,U,S,D,B> implements VariableBindingMapper<L,U,S,D,B> {

  /**
   * Format a Long value as a string according to the given DISPLAY-HINT format.
   *
   * @param value
   *    the long value to format
   * @param displayHint
   *    the format description as defined by SMIv2 for integer values.
   * @return
   *    the formatted string representation of the value.
   */
  public static String formatLong(Long value, String displayHint) {
    String s = null;
    if ((displayHint == null) || (displayHint.length() < 1)) return null;
    char c = displayHint.charAt(0);
    switch (c) {
      case 'd': {
        int dec = -1;
        if ((displayHint.length() > 2) && (displayHint.charAt(1) == '-')) {
          dec = Integer.parseInt(displayHint.substring(2));
        }
        s = value.toString();
        if (dec > 0) {
          int pos = s.length()-dec;
          if (pos <= 0)
            s = "."+s;
          else
            s = s.substring(0, pos)+"."+s.substring(pos);
        }
        break;
      }
      case 'b': {
        s = Long.toBinaryString(value);
        break;
      }
      case 'o': {
        s = Long.toOctalString(value);
        break;
      }
      case 'x': {
        s = Long.toHexString(value);
        break;
      }
    }
    return s;
  }

  /**
   * Formats an {@link OctetString} value using a DISPLAY-HINT format to a String value.
   * @param value
   *    an {@link OctetString} instance.
   * @param displayHint
   *    a SMIv2 DISPLAY-HINT string format specification.
   * @return
   *    a String representation of the <code>value</code>.
   */
  public static String formatString(OctetString value, String displayHint) {
    if (displayHint == null) return value.toString();
    OctetString output = new OctetString();
    int pd = 0;
    int pv = 0;

    while ((pv < value.length()) && (pd < displayHint.length())) {

      int repetition = 1;
      Character sep = null;
      Character repTerm = null;

      if (displayHint.charAt(pd) == '*') {
        pd++;
        repetition = 0xFF & value.get(pv++);
      }
      int end = pd;
      while ((end < displayHint.length()) &&
             (Character.isDigit(displayHint.charAt(end)))) {
        end++;
      }
      if (end == pd) {
        return value.toString(); // format error
      }
      String str = null;
      if (end < displayHint.length()) {
        str = displayHint.substring(pd, end);
      }
      else {
        return value.toString(); // format error
      }

      int n = Math.min(Integer.parseInt(str),65535);
      pd = end;
      char mode = displayHint.charAt(pd++);
      if ((pd < displayHint.length()) &&
          (!Character.isDigit(displayHint.charAt(pd)) && (displayHint.charAt(pd) != '*'))) {
        sep = displayHint.charAt(pd);
        pd++;
        if ((pd < displayHint.length()) && (!Character.isDigit(displayHint.charAt(pd)) &&
            (displayHint.charAt(pd) != '*'))) {
          repTerm = displayHint.charAt(pd);
          pd++;
        }
      }
      for (int k=0; k<repetition; k++) {
        if (pv+n >= value.length()) {
          n = value.length()-pv;
        }
        if (n > 0) {
          if (pv >= value.length()) {
            break;
          }
          byte[] val = new byte[n];
          for (int z=0; z<n; z++) {
            val[z] = value.get(pv+z);
          }
          pv += n;
          if ((mode == 'a') || (mode == 't')) {
            output.append(val);
          }
          else {
            long l = 0;
            for (byte aVal : val) {
              l = l * 256 + (aVal & 0xFF);
            }
            switch (mode) {
              case 'd': {
                output.append(Long.toString(l));
                break;
              }
              case 'b': {
                output.append(Long.toBinaryString(l));
                break;
              }
              case 'o': {
                output.append(Long.toOctalString(l));
                break;
              }
              case 'x': {
                String h = Long.toHexString(l);
                if (h.length() == 1)
                  h = "0"+h;
                output.append(h);
                break;
              }
            }
          }
        }
        else {
          break;
        }
        if ((sep != null) && (pv < value.length())) {
          output.append((byte)sep.charValue());
        }
        if (k+1 < repetition) {
          output.append((byte)repTerm.charValue());
        }
        if (pd >= displayHint.length()) {
          pd = 0;
        }
      }
    }
    return new String(output.getValue());
  }

  /**
   * Scans a String representation of an {@link OctetString} using a DISPLAY-HINT format.
   * @param format
   *    the SMIv2 DISPLAY-HINT format.
   * @param value
   *    a String representation of an {@link OctetString} that matches the given format.
   * @return
   *    the {@link OctetString} representation of the given value.
   */
  public static OctetString scanString(String format, String value) {
    if (format == null) return new OctetString(value);
    int i = 0;
    boolean valid;
    int fmt = 0;
    int last;
    boolean have_pfx;
    int pfx;
    int vv = 0;

    if ((value == null) || (format == null)) return null;
    OctetString out = new OctetString();
    while ((fmt < format.length()) && (i<value.length())) {
      last = fmt;
      have_pfx = false;
      pfx = 0;
      while ((fmt < format.length()) && (Character.isDigit(format.charAt(fmt)))) {
        pfx = pfx * 10 + format.charAt(fmt) - '0';
        have_pfx = true;
        fmt++;
      }
      if (!have_pfx) {
        pfx = 1;
      }
      valid = false;
      switch (format.charAt(fmt)) {
        case 't':
        case 'a': {
          if (pfx < (value.length() - i)) {
            out.append(value.substring(i, i+pfx).getBytes());
            i += pfx;
          }
          else {
            out.append(value.substring(i).getBytes());
            i = value.length();
          }
          break;
        }
        case 'b': {
          int start = i;
          while ((i<value.length()) && ((value.charAt(i) == '0') || (value.charAt(i) == '1'))) {
            i++;
          }
          try {
            vv = Integer.parseInt(value.substring(start, i), 2);
            valid = true;
          }
          catch (NumberFormatException nfe) {
            valid = false;
          }
          break;
        }
        case 'd': {
          int start = i;
          while ((i<value.length()) && (Character.isDigit(value.charAt(i)))) {
            i++;
          }
          try {
            vv = Integer.parseInt(value.substring(start, i), 10);
            valid = true;
          }
          catch (NumberFormatException nfe) {
            valid = false;
          }
          break;
        }
        case 'o': {
          int start = i;
          while ((i<value.length()) &&
              (value.charAt(i)>='0') &&
              (value.charAt(i)<='7')) {
            i++;
          }
          try {
            vv = Integer.parseInt(value.substring(start, i), 8);
            valid = true;
          }
          catch (NumberFormatException nfe) {
            valid = false;
          }
          break;
        }
        case 'x': {
          int start = i;
          while ((i<value.length()) &&
              (((value.charAt(i)>='0') &&
                  (value.charAt(i)<='9')) ||
                  ((value.charAt(i)>='A') &&
                      (value.charAt(i)<='F')) ||
                  ((value.charAt(i)>='a') &&
                      (value.charAt(i)<='f')))) {
            i++;
          }
          try {
            vv = Integer.parseInt(value.substring(start, i), 16);
            valid = true;
          }
          catch (NumberFormatException nfe) {
            valid = false;
          }
          break;
        }
        default: {
          return null;
        }
      }
      fmt++;
      if (valid) {
        while (pfx > 0) {
          byte c = (byte)((vv >> ((pfx -1)*8)) & 0xFF);
          out.append(c);
          pfx--;
        }
      }
      // Check for separator and repeat with the last format if
      // there is still data available.
      if ((fmt < format.length()) && (format.charAt(fmt) != '*')) {
        // skip over format separator
        if (!Character.isDigit(format.charAt(fmt))) {
          fmt++;
        }
        // skip over data separator
        if ((i<value.length()) && (!Character.isDigit(value.charAt(i)))) {
          i++;
        }
      }
      if ((fmt >= format.length()) && (i<value.length())) {
        if (valid) {
          fmt = last;
        }
        else {
          return null;
        }
      }
    }
    return out;
  }

  /**
   * Scans a String value as a Long using the specified SMIv2 DISPLAY-HINT format specification.
   * @param format
   *    the SMIv2 DISPLAY-HINT format.
   * @param value
   *    a string value representing a number with the DISPLAY-HINT <code>format</code>.
   * @return
   *    the Long representation of the given value or <code>null</code> if the value could not by
   *    scanned using the given format.
   * @throws NumberFormatException
   *    if value and format render a string that cannot be parsed as a long value.
   */
  public static Long scanLong(String format, String value)
      throws NumberFormatException
  {
    if (format == null) {
      return new Long(value);
    }
    Long out = null;
    long l = 0;
    int v = 0;
    int dpt = 0;
    int i = 0;
    boolean sign = false;

    if ((format.length() < 1) || (value == null) || (value.length() < 1)) {
      return null;
    }

    switch (format.charAt(0)) {
      case 'd': {
        if (format.length() == 1) {
          out = new Long(value);
        }
        else {
          if (format.length() < 2) return null;
          if (format.charAt(1) != '-') break; // invalid format string
          if (Character.isDigit(format.charAt(2))) {
            for (dpt = 0, i=0; ((i+2<format.length()) &&
                (Character.isDigit(format.charAt(2+i)))); i++) {
              dpt = dpt * 10 + (format.charAt(2+i) - '0');
            }
          }
          if (2+i < format.length()) break; // invalid format string
          if (value.charAt(0) == '-') {
            sign = true;
            v++;
          }
          int frac = -1;
          for (l = 0; ((v < value.length()) &&
              ((Character.isDigit(value.charAt(v))) ||
                  (value.charAt(v) == '.'))); v++) {
            if (value.charAt(v) == '.') {
              if (frac >= 0) break;
              frac = 0;
              continue;
            }
            l = l * 10 + (value.charAt(v) - '0');
            if (frac >= 0) {
              frac++;
            }
          }
          if (v >= value.length()) {
            for (; frac < dpt; frac++) l *= 10;
            for (; frac > dpt; frac--) l /= 10;
            out = (sign ? -1 * l : l);
          }
        }
        break;
      }
      case 'x': {
        if (format.length() > 1) break; // invalid format string
        out = Long.parseLong(value, 16);
        break;
      }
      case 'o': {
        if (format.length() > 1) break; // invalid format string
        out = Long.parseLong(value, 8);
        break;
      }
      case 'b': {
        if (format.length() > 1) break; // invalid format string
        out = Long.parseLong(value, 2);
        break;
      }
      default:
        break;
    }
    return out;
  }


  /**
   * Creates a DateAndTime <code>OctetString</code> value from a
   * <code>GregorianCalendar</code>.
   * @param dateAndTime
   *    a <code>GregorianCalendar</code> instance.
   * @return
   *    the {@link OctetString} instance with the set calendar value.
   */
  public static OctetString makeDateAndTime(GregorianCalendar dateAndTime) {
    int length = 0;
    byte[] bytes = new byte[11];
    bytes[length++] = (byte) (dateAndTime.get(Calendar.YEAR) / 256);
    bytes[length++] = (byte) (dateAndTime.get(Calendar.YEAR) % 256);
    bytes[length++] = (byte) (dateAndTime.get(Calendar.MONTH) + 1);
    bytes[length++] = (byte) (dateAndTime.get(Calendar.DAY_OF_MONTH));
    bytes[length++] = (byte) (dateAndTime.get(Calendar.HOUR_OF_DAY));
    bytes[length++] = (byte) (dateAndTime.get(Calendar.MINUTE));
    bytes[length++] = (byte) (dateAndTime.get(Calendar.SECOND));
    bytes[length++] = (byte) (dateAndTime.get(Calendar.MILLISECOND) / 100);
    if (dateAndTime.getTimeZone() != null) {
      TimeZone tz = dateAndTime.getTimeZone();
      bytes[length++] = (byte) ((tz.getRawOffset() >= 0) ? '+' : '-');
      bytes[length++] = (byte) (tz.getOffset(dateAndTime.getTimeInMillis()) / 3600000);
      bytes[length++] = (byte) ((tz.getOffset(dateAndTime.getTimeInMillis()) % 3600000) / 60000);
    }
    return new OctetString(bytes, 0, length);
  }

  /**
   * Creates a <code>GregorianCalendar</code> from a properly formatted
   * DateAndTime <code>OctetString</code>.
   * @param dateAndTimeValue
   *    an OctetString conforming to the DateAndTime TC.
   * @return
   *    the corresponding <code>GregorianCalendar</code> instance.
   */
  /**
   * Creates a <code>GregorianCalendar</code> from a properly formatted
   * DateAndTime <code>OctetString</code>.
   * @param dateAndTimeValue
   *    an OctetString conforming to the DateAndTime TC.
   * @return
   *    the corresponding <code>GregorianCalendar</code> instance.
   */
  public static GregorianCalendar makeCalendar(OctetString dateAndTimeValue) {
    int year = (dateAndTimeValue.get(0) & 0xFF ) * 256 +
        (dateAndTimeValue.get(1) & 0xFF );
    int month = (dateAndTimeValue.get(2) & 0xFF );
    int date = (dateAndTimeValue.get(3) & 0xFF );
    int hour = (dateAndTimeValue.get(4) & 0xFF );
    int minute = (dateAndTimeValue.get(5) & 0xFF );
    int second = (dateAndTimeValue.get(6) & 0xFF );
    int deci = (dateAndTimeValue.get(7) & 0xFF );
    if (dateAndTimeValue.length() == 11) {
      String timezone = String.format("GMT%c%02d:%02d", dateAndTimeValue.get(8),
          dateAndTimeValue.get(9), dateAndTimeValue.get(10));
      GregorianCalendar tgc =
          new GregorianCalendar(TimeZone.getTimeZone(timezone));
      tgc.set(year, month-1, date, hour, minute, second);
      tgc.set(Calendar.MILLISECOND, deci*100);
      return tgc;
    }
    else {
      GregorianCalendar gc =
          new GregorianCalendar(year, month-1, date, hour, minute, second);
      gc.set(Calendar.MILLISECOND, deci*100);
      return gc;
    }
  }

  /**
   * Tests a variable for DateAndTime conformance.
   * @param dateAndTime
   *    a Variable.
   * @return
   *    {@link org.snmp4j.mp.SnmpConstants#SNMP_ERROR_SUCCESS} if <code>dateAndTime</code>
   *    is valid or an appropriate SNMP error code if not.
   */
  public static int validateDateAndTime(Variable dateAndTime) {
    if (dateAndTime instanceof OctetString) {
      OctetString os = (OctetString)dateAndTime;
      if ((os.length() != 8) && (os.length() != 11)) {
        return SnmpConstants.SNMP_ERROR_WRONG_LENGTH;
      }
      int month = (os.get(2) & 0xFF );
      int date = (os.get(3) & 0xFF );
      int hour = (os.get(4) & 0xFF );
      int minute = (os.get(5) & 0xFF );
      int second = (os.get(6) & 0xFF );
      int deci = (os.get(7) & 0xFF );
      if ((month < 1) || (month > 12) ||
          (date < 1) || (date > 31) || (hour > 23) || (second > 59) ||
          (minute > 59) || (deci > 9)) {
        return SnmpConstants.SNMP_ERROR_WRONG_VALUE;
      }
      if (os.length() == 11) {
        if ((os.get(8) != '+') && (os.get(8) != '-')) {
          return SnmpConstants.SNMP_ERROR_WRONG_VALUE;
        }
      }
      return SnmpConstants.SNMP_ERROR_SUCCESS;
    }
    return SnmpConstants.SNMP_ERROR_WRONG_TYPE;
  }

}
