/*_############################################################################
  _## 
  _##  SNMP4J-Model - VariableBindingMapperJava.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/

package org.snmp4j.model.mapper;

import org.snmp4j.smi.*;

import java.math.BigInteger;
import java.text.ParseException;
import java.util.Date;
import java.util.GregorianCalendar;

/**
 * <code>VariableBindingMapperJava</code> maps SNMP4J {@link VariableBinding} variables
 * to Java native language objects and vice versa.
 *
 * @author Frank Fock
 * @version 1.0
 */
public class VariableBindingMapperJava
    extends AbstractVariableBindingMapper<Long,BigInteger,String,GregorianCalendar,byte[]> {

  @Override
  public Long toLong(VariableBinding variableBinding) {
    return variableBinding.getVariable().toLong();
  }

  @Override
  public BigInteger toUnsignedLong(VariableBinding variableBinding) {
    if (variableBinding.getVariable() instanceof Counter64) {
      byte[] bytes = new byte[8];
      for (int i = 0; i < 8; i++) {
        bytes[i] = (byte) ((((Counter64) variableBinding.getVariable()).getValue() >> ((7 - i) * 8)) & 0xFF);
      }
      return new BigInteger(1, bytes);
    }
    return new BigInteger(variableBinding.getVariable().toString());
  }

  @Override
  public String toString(VariableBinding variableBinding) {
    return variableBinding.toValueString();
  }

  @Override
  public GregorianCalendar toDate(VariableBinding variableBinding) {
    if (variableBinding.getVariable() instanceof OctetString) {
      return makeCalendar((OctetString) variableBinding.getVariable());
    }
    throw new IllegalArgumentException("Only OctetString with DateAndTime TC can be converted to GregorianCalendar: "+
        variableBinding);
  }

  @Override
  public byte[] toBinary(VariableBinding variableBinding) {
    if (variableBinding.getVariable() instanceof OctetString) {
      return ((OctetString) variableBinding.getVariable()).getValue();
    }
    throw new IllegalArgumentException("Only OctetString can be converted to an byte array: " + variableBinding);
  }

  @Override
  public Object toObject(VariableBinding variableBinding, Class targetClass) {
    if (targetClass.equals(GregorianCalendar.class)) {
      return toDate(variableBinding);
    }
    else if (targetClass.equals(String.class)) {
      return toString(variableBinding);
    }
    else if (targetClass.equals(byte[].class)) {
      return toBinary(variableBinding);
    }
    else if (targetClass.equals(Long.class)) {
      return toLong(variableBinding);
    }
    else if (BigInteger.class.equals(targetClass)) {
      return toUnsignedLong(variableBinding);
    }
    return null;
  }

  @Override
  public VariableBinding fromLong(OID oid, Long longValue) {
    try {
      return new VariableBinding(oid, longValue.toString());
    } catch (ParseException e) {
      throw new IllegalArgumentException("The OID "+oid+" cannot be bound to a long value: "+e.getMessage(), e);
    }
  }

  @Override
  public VariableBinding fromUnsignedLong(OID oid, BigInteger unsignedLongValue) {
    try {
      return new VariableBinding(oid, unsignedLongValue.toString());
    } catch (ParseException e) {
      throw new IllegalArgumentException("The OID "+oid+" cannot be bound to an unsigned long value: "+e.getMessage(),
          e);
    }
  }

  @Override
  public VariableBinding fromString(OID oid, String stringValue) {
    try {
      return new VariableBinding(oid, stringValue);
    } catch (ParseException e) {
      throw new IllegalArgumentException("The OID "+oid+" cannot be bound to a string value: "+e.getMessage(), e);
    }
  }

  @Override
  public VariableBinding fromDate(OID oid, GregorianCalendar dateValue) {
    return new VariableBinding(oid, makeDateAndTime(dateValue));
  }

  @Override
  public VariableBinding fromBinary(OID oid, byte[] binaryValue) {
    return new VariableBinding(oid, new OctetString(binaryValue));
  }

  @Override
  public VariableBinding fromObject(OID oid, Object object) {
    if (object instanceof byte[]) {
      return fromBinary(oid, (byte[])object);
    }
    else if (object instanceof GregorianCalendar) {
      return fromDate(oid, (GregorianCalendar) object);
    }
    else if (object instanceof BigInteger) {
      return fromUnsignedLong(oid, (BigInteger) object);
    }
    else if (object instanceof Long) {
      return fromLong(oid, (Long) object);
    }
    else if (object == null) {
      return new VariableBinding(oid);
    }
    String s = object.toString();
    if (s == null) {
      return new VariableBinding(oid);
    }
    return fromString(oid, object.toString());
  }
}
