/*_############################################################################
  _## 
  _##  SNMP4J-Model - SNMP4JAgentSnmpService.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.agent;

import org.snmp4j.PDU;
import org.snmp4j.Target;
import org.snmp4j.agent.*;
import org.snmp4j.agent.mo.*;
import org.snmp4j.agent.request.*;
import org.snmp4j.log.LogAdapter;
import org.snmp4j.log.LogFactory;
import org.snmp4j.model.snmp.proxy.SnmpCommitListener;
import org.snmp4j.model.snmp.proxy.impl.SnmpValuesChangeSet;
import org.snmp4j.model.snmp.spi.*;
import org.snmp4j.model.snmp.api.*;
import org.snmp4j.model.transaction.SnmpIsolationLevel;
import org.snmp4j.model.transaction.SnmpTransaction;
import org.snmp4j.model.transaction.SnmpTransactionStrategy;
import org.snmp4j.smi.*;

import java.io.IOException;
import java.util.*;

/**
 * The SNMP4JAgentSnmpService implements the {@link org.snmp4j.model.snmp.spi.SnmpService} interface for
 * SNMP4J-Agent agents. Instead of communication using UDP or TCP transport, this service directly provides
 * access to a {@link MOServer} instance.
 */
public class SNMP4JAgentSnmpService implements SnmpService {

  private static LogAdapter LOG = LogFactory.getLogger(SNMP4JAgentSnmpService.class);

  private MOServer[] moServers;
  private static final OctetString DEFAULT_CONTEXT = new OctetString();
  private WeakHashMap<GlobalTransaction, Map<Target, LocalTransaction>> transactions =
      new WeakHashMap<GlobalTransaction, Map<Target, LocalTransaction>>();

  /**
   * Create a {@link SnmpService} based on an array of managed object servers.
   * @param moServers
   *    an array of {@link MOServer} instances that serve the managed objects available for this SNMP service.
   */
  public SNMP4JAgentSnmpService(MOServer[] moServers) {
    this.moServers = moServers;
  }

  @Override
  public void initTransports() throws IOException {

  }

  @Override
  public List<? extends VariableBinding> get(Target target, List<OID> oidList) throws SnmpException {
    OctetString context = DEFAULT_CONTEXT;
    MOServer moServer = getServer(context);
    LOG.debug("GET on target "+target+" OIDs="+oidList);
    ArrayList<VariableBinding> vbs = new ArrayList<VariableBinding>();
    for (OID oid : oidList) {
      MOContextScope scope =new DefaultMOContextScope(context, oid, true, oid, true);
      MOQuery moQuery = new DefaultMOQuery(scope, false);
      ManagedObject mo = moServer.lookup(moQuery);
      LOG.debug("Got MO = "+mo+" for query "+moQuery);
      if (mo != null) {
        SubRequest subRequest = new AgentSubRequest(scope, moQuery, new VariableBinding(oid), 0);
        mo.get(subRequest);
        LOG.debug("GET SubRequest returned "+subRequest+" for OID = "+oid);
        if (subRequest.hasError()) {
          throw new SnmpException(SnmpErrorStatus.fromSnmpErrorStatus(subRequest.getErrorStatus()));
        }
        vbs.add(subRequest.getVariableBinding());
      }
    }
    LOG.debug("Returning GET VBS = "+vbs);
    return vbs;
  }

  @Override
  public List<? extends VariableBinding> getNext(Target target, List<OID> oidList) throws SnmpException {
    OctetString context = DEFAULT_CONTEXT;
    MOServer moServer = getServer(context);
    LOG.debug("GETNEXT on target "+target+" OIDs="+oidList);
    ArrayList<VariableBinding> vbs = new ArrayList<VariableBinding>();
    for (OID oid : oidList) {
      MOContextScope scope = new DefaultMOContextScope(context, oid, false, oid, true);
      MOQuery moQuery = new DefaultMOQuery(scope, false);
      ManagedObject mo = moServer.lookup(moQuery);
      LOG.debug("Got MO = "+mo+" for query "+moQuery);
      if (mo != null) {
        SubRequest subRequest = new AgentSubRequest(scope, moQuery, new VariableBinding(oid), 0);
        mo.next(subRequest);
        LOG.debug("NEXT SubRequest returned " + subRequest + " for OID = " + oid);
        if (subRequest.hasError()) {
          throw new SnmpException(SnmpErrorStatus.fromSnmpErrorStatus(subRequest.getErrorStatus()));
        }
        vbs.add(subRequest.getVariableBinding());
      }
    }
    LOG.debug("Returning NEXT VBS = "+vbs);
    return vbs;
  }

  @Override
  public List<? extends VariableBinding> getSubTree(Target target, OID oid) throws SnmpException {
    OctetString context = DEFAULT_CONTEXT;
    MOServer moServer = getServer(context);
    LOG.debug("SUBTREE on target "+target+" OIDs="+oid);
    ArrayList<VariableBinding> vbs = new ArrayList<VariableBinding>();
    OID lastOID = oid;
    while (lastOID.startsWith(oid)) {
      MOContextScope scope = new DefaultMOContextScope(context, lastOID, false, oid.nextPeer(), false);
      MOQuery moQuery = new DefaultMOQuery(scope, false);
      ManagedObject mo = moServer.lookup(moQuery);
      LOG.debug("Got MO = "+mo+" for query "+moQuery);
      if (mo != null) {
        SubRequest subRequest = new AgentSubRequest(scope, moQuery, new VariableBinding(lastOID), 0);
        mo.next(subRequest);
        LOG.debug("NEXT SubRequest returned " + subRequest + " for OID = " + lastOID);
        if (subRequest.hasError()) {
          throw new SnmpException(SnmpErrorStatus.fromSnmpErrorStatus(subRequest.getErrorStatus()));
        }
        if (subRequest.getVariableBinding().getOid().compareTo(lastOID) > 0) {
          vbs.add(subRequest.getVariableBinding());
          lastOID = subRequest.getVariableBinding().getOid();
        }
        else {
          break;
        }
      }
      else {
        break;
      }
    }
    LOG.debug("Returning SUBTREE VBS = "+vbs);
    return vbs;
  }

  @Override
  public List<? extends SnmpRow> getTable(Target target, List<OID> columnOIDs, OID minIndex, OID maxIndex) throws SnmpException {
    OctetString context = DEFAULT_CONTEXT;
    MOServer moServer = getServer(context);
    LOG.debug("TABLE on target "+target+" OIDs="+columnOIDs+", minIndex = "+minIndex+ ", maxIndex = "+maxIndex);
    TreeMap<OID, SnmpRow> cellBuffer = new TreeMap<OID, SnmpRow>();
    Set<ManagedObject> visitedMOs = new HashSet<ManagedObject>();
    for (int c = 0; c<columnOIDs.size(); c++) {
      OID column = columnOIDs.get(c);
      OID startCell = (minIndex != null) ? new OID(column.getValue(), minIndex.getValue()) : column;
      OID endCell = (maxIndex != null) ? new OID(column.getValue(), maxIndex.getValue()) : column.successor();
      MOContextScope scope = new DefaultMOContextScope(context, startCell, (minIndex != null), endCell, maxIndex != null);
      MOQuery moQuery = new DefaultMOQuery(scope, false);
      ManagedObject mo = moServer.lookup(moQuery);
      if ((mo != null) && (!visitedMOs.contains(mo))) {
        LOG.debug("Found MO = "+mo+" for table query "+moQuery);
        visitedMOs.add(mo);
        if (mo instanceof MOTable) {
          Iterator<MOTableRow> rowIterator = ((MOTable) mo).getModel().tailIterator(minIndex);
          while (rowIterator.hasNext()) {
            MOTableRow row = rowIterator.next();
            LOG.debug("ROW returned: "+row);
            if (maxIndex == null || row.getIndex().compareTo(maxIndex) <= 0) {
              List<VariableBinding> columnVbs = Arrays.asList(new VariableBinding[columnOIDs.size()]);
              SnmpRow snmpRow = cellBuffer.get(row.getIndex());
              if (snmpRow == null) {
                snmpRow = new SnmpRow(row.getIndex(), columnVbs);
              }
              for (int cr = c; cr < columnOIDs.size(); cr++) {
                if (columnOIDs.get(cr).startsWith(((MOTable)mo).getOID())) {
                  for (int i = 0; i<row.size() && i<((MOTable) mo).getColumnCount(); i++) {
                    if (columnOIDs.get(cr).last() == ((MOTable)mo).getColumn(i).getColumnID()) {
                      Variable variable = row.getValue(i);
                      if (variable == null) {
                        variable = Null.noSuchInstance;
                      }
                      columnVbs.set(cr, new VariableBinding(new OID(columnOIDs.get(cr).getValue(),
                          snmpRow.getIndex().getValue()), variable));
                    }
                  }
                }
              }
              LOG.debug("Putting ROW in cell buffer: "+snmpRow);
              cellBuffer.put(snmpRow.getIndex(), snmpRow);
            }
          }
        }
        else {
          List<? extends VariableBinding> columnValues = getSubTree(target, startCell);
          for (VariableBinding vb : columnValues) {
            OID index = new OID(vb.getOid().getValue(), column.size(), vb.getOid().size() - column.size());
            SnmpRow row = cellBuffer.get(index);
            if (row == null) {
              List<VariableBinding> columnVbs = Arrays.asList(new VariableBinding[columnOIDs.size()]);
              row = new SnmpRow(index, columnVbs);
              LOG.debug("Putting ROW in cell buffer: "+row);
              cellBuffer.put(index, row);
              columnVbs.set(c, vb);
            }
            else {
              LOG.debug("Setting cell col="+c+" to vb="+vb);
              row.getVariableBindings().set(c, vb);
            }
          }
        }
      }
    }
    LOG.debug("Returning ROWS = "+cellBuffer.values());
    return new ArrayList<SnmpRow>(cellBuffer.values());
  }

  @Override
  public void getTable(Target target, List<OID> columnOIDs, OID minIndex, OID maxIndex, SnmpRowListener callback, Object userObject) throws SnmpException {
    List<? extends SnmpRow> rows = getTable(target, columnOIDs, minIndex, maxIndex);
    callback.beginUpdate(userObject);
    for (SnmpRow row : rows) {
      callback.nextRow(row, userObject);
    }
    callback.endUpdate(SnmpErrorStatus.noError, userObject);
  }

  @Override
  public void update(SnmpTransaction snmpTransaction, Target target, SnmpValuesChangeSet variableBindings,
                     SnmpCommitListener snmpCommitListener) {
    Map<Target,LocalTransaction> localTransactions = transactions.get(snmpTransaction);
    if (localTransactions == null) {
      throw new IllegalStateException("Transaction not available: "+snmpTransaction);
    }
    LocalTransaction localTransaction = localTransactions.get(target);
    if (localTransaction == null) {
      localTransaction = new LocalTransaction(target);
      localTransactions.put(target, localTransaction);
    }
    localTransaction.addUpdate(variableBindings);
    if (snmpTransaction.isAutoCommit()) {
      commit(snmpTransaction, snmpCommitListener);
    }
  }

  @Override
  public List<? extends SnmpCommitResult> commit(SnmpTransaction transaction, SnmpCommitListener commitListener) {
    OctetString context = DEFAULT_CONTEXT;
    MOServer moServer = getServer(context);
    Map<Target,LocalTransaction> localTransactions = transactions.get(transaction);
    if (localTransactions == null) {
      throw new IllegalStateException("Transaction not available: "+transaction);
    }
    ArrayList<SnmpCommitResult> commitResults = new ArrayList<SnmpCommitResult>();
    for (Map.Entry<Target, LocalTransaction> localTransactionEntry : localTransactions.entrySet()) {
      LocalTransaction localTransaction = localTransactionEntry.getValue();
      Queue<SnmpValuesChangeSet> updates = localTransaction.getPending();
      List<SnmpValuesChangeSet> localTransactionPending =
          new ArrayList<SnmpValuesChangeSet>();
      List<VariableBinding> chunk = new ArrayList<VariableBinding>();
      while ((!updates.isEmpty()) && localTransaction.getFailedVariableBindings().isEmpty()) {
        SnmpValuesChangeSet nextUpdate = updates.poll();
        chunk.addAll(nextUpdate.getVariableBindings());
        localTransactionPending.add(nextUpdate);
      }
      SnmpCommitResult result = null;
      for (VariableBinding vb : chunk) {
        MOContextScope scope = new DefaultMOContextScope(context, vb.getOid(), true, vb.getOid(), true);
        MOQuery moQuery = new DefaultMOQuery(scope, true);
        ManagedObject mo =  moServer.lookup(moQuery);
        LOG.debug("Committing VB = "+vb+" to "+mo+" return by query "+moQuery);
        if (mo != null) {
          if (mo instanceof MOTable && ((MOTable)mo).getModel() instanceof MOMutableTableModel) {
            MOTable moTable = (MOTable)mo;
            MOMutableTableModel tableModel = (MOMutableTableModel)moTable.getModel();
            MOTableCellInfo cellInfo = moTable.getCellInfo(vb.getOid());
            OID rowIndex = cellInfo.getIndex();
            MOTableRow row = moTable.getModel().getRow(rowIndex);
            if (row == null) {
              row = tableModel.createRow(rowIndex, moTable.getDefaultValues());
              moTable.addRow(row);
            }
            if (row instanceof MOMutableTableRow) {
              MOMutableTableRow mutableTableRow = (MOMutableTableRow)row;
              mutableTableRow.setValue(cellInfo.getColumn(), (Variable)vb.getVariable().clone());
              LOG.debug("Committed VB = " + vb + " to row " + row + " at " + cellInfo);
            }
          }
          else if (mo instanceof MOScalar) {
            MOScalar moScalar = (MOScalar)mo;
            LOG.debug("Committing VB = "+vb+" to scalar "+moScalar);
            moScalar.setValue(vb.getVariable());
          }
        }
      }
      /*
      //prepare
      ArrayList<SubRequest> subRequests = new ArrayList<SubRequest>();
      for (VariableBinding vb : chunk) {
        MOContextScope scope = new DefaultMOContextScope(context, vb.getOid(), true, vb.getOid(), true);
        MOQuery moQuery = new DefaultMOQuery(scope, true);
        ManagedObject mo =  moServer.lookup(moQuery);
        if (mo != null) {
          SubRequest subRequest = new AgentSubRequest(scope, moQuery, vb, i);
          subRequest.setTargetMO(mo);
          subRequests.add(subRequest);
          mo.prepare(subRequest);
          if (subRequest.hasError()) {
            phase = Request.PHASE_2PC_CLEANUP;
            result = new CommitStatus(SnmpErrorStatus.fromSnmpErrorStatus(subRequest.getErrorStatus()),
                subRequest.getIndex(), Collections.EMPTY_LIST, localTransactionPending);
            break;
          }
        }
        else {
          phase = Request.PHASE_2PC_CLEANUP;
          result = new CommitStatus(SnmpErrorStatus.noSuchName, i, Collections.<SnmpValuesChangeSet>emptyList(),
              localTransactionPending);
          break;
        }
        i++;
      }
      i=0;
      if (phase != Request.PHASE_2PC_CLEANUP) {
        // commit
        for (SubRequest subRequest : subRequests) {
          ManagedObject mo = subRequest.getTargetMO();
          mo.commit(subRequest);
          if (subRequest.hasError()) {
            phase = Request.PHASE_2PC_UNDO;
            result = new CommitStatus(SnmpErrorStatus.fromSnmpErrorStatus(subRequest.getErrorStatus()),
                subRequest.getIndex(), Collections.<SnmpValuesChangeSet>emptyList(), localTransactionPending);
            break;
          }
          i++;
        }
      }
      if (phase == Request.PHASE_2PC_CLEANUP) {
        // cleanup
        for (SubRequest subRequest : subRequests) {
          ManagedObject mo = subRequest.getTargetMO();
          mo.cleanup(subRequest);
        }
      }
      else if (phase == Request.PHASE_2PC_UNDO) {
        // cleanup
        for (SubRequest subRequest : subRequests) {
          ManagedObject mo = subRequest.getTargetMO();
          mo.undo(subRequest);
        }
      }
      */
      if (result == null) {
        result = new CommitStatus(SnmpErrorStatus.noError, 0,
            localTransactionPending, Collections.<SnmpValuesChangeSet>emptyList());
      }
      commitResults.add(result);
    }
    return commitResults;
  }

  @Override
  public void rollback(SnmpTransaction transaction) {
    Map<Target,LocalTransaction> localTransactions = transactions.get(transaction);
    if (localTransactions != null) {
      for (LocalTransaction localTransaction : localTransactions.values()) {
        localTransaction.getPending().clear();
      }
    }
  }

  @Override
  public SnmpTransaction getSnmpTransaction(SnmpTransactionStrategy transactionStrategy) {
    GlobalTransaction globalTransaction =
        new GlobalTransaction(SnmpIsolationLevel.none == transactionStrategy.getIsolationLevel());
    transactions.put(globalTransaction, new LinkedHashMap<Target, LocalTransaction>());
    return globalTransaction;
  }

  @Override
  public boolean isAtomic(SnmpTransaction transaction) {
    return false;
  }

  private MOServer getServer(OctetString context) {
    for (MOServer moServer : moServers) {
      if (moServer.isContextSupported(context)) {
        return moServer;
      }
    }
    return null;
  }

  private class AgentSubRequest implements SubRequest {

    private RequestStatus requestStatus;
    private MOScope moScope;
    private VariableBinding variableBinding;
    private Object undoValue;
    private ManagedObject targetMO;
    private MOQuery moQuery;
    private int index;
    private volatile Object userObject;

    private AgentSubRequest(MOScope moScope, MOQuery moQuery, VariableBinding variableBinding, int index) {
      this.moScope = moScope;
      this.variableBinding = variableBinding;
      this.moQuery = moQuery;
      this.index = index;
      this.requestStatus = new RequestStatus();
    }

    @Override
    public boolean hasError() {
      return requestStatus.getErrorStatus() != 0;
    }

    @Override
    public void setErrorStatus(int errorStatus) {
      this.requestStatus.setErrorStatus(errorStatus);
    }

    @Override
    public int getErrorStatus() {
      return this.requestStatus.getErrorStatus();
    }

    @Override
    public RequestStatus getStatus() {
      return requestStatus;
    }

    @Override
    public MOScope getScope() {
      return moScope;
    }

    @Override
    public VariableBinding getVariableBinding() {
      return variableBinding;
    }

    @Override
    public Request getRequest() {
      return null;
    }

    @Override
    public Object getUndoValue() {
      return undoValue;
    }

    @Override
    public void setUndoValue(Object undoInformation) {
       this.undoValue = undoInformation;
    }

    @Override
    public void completed() {
      requestStatus.setPhaseComplete(true);
    }

    @Override
    public boolean isComplete() {
      return requestStatus.isPhaseComplete();
    }

    @Override
    public void setTargetMO(ManagedObject managedObject) {
       this.targetMO = managedObject;
    }

    @Override
    public ManagedObject getTargetMO() {
      return targetMO;
    }

    @Override
    public int getIndex() {
      return index;
    }

    @Override
    public void setQuery(MOQuery query) {
      this.moQuery = query;
    }

    @Override
    public MOQuery getQuery() {
      return moQuery;
    }

    @Override
    public SubRequestIterator<? extends SubRequest> repetitions() {
      return null;
    }

    @Override
    public void updateNextRepetition() {

    }

    @Override
    public Object getUserObject() {
      return userObject;
    }

    @Override
    public void setUserObject(Object userObject) {
      this.userObject = userObject;
    }
  }

}
