/*_############################################################################
  _## 
  _##  SNMP4J-Model - TableHelper.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.api;

import com.snmp4j.smi.*;
import org.snmp4j.model.demo.mibbrowser.SimpleMibBrowser;
import org.snmp4j.model.snmp.proxy.SnmpProxyColumn;
import org.snmp4j.model.snmp.proxy.SnmpProxyColumnFactory;
import org.snmp4j.model.snmp.proxy.SnmpProxyIndexColumn;
import org.snmp4j.model.snmp.proxy.impl.SnmpColumn;
import org.snmp4j.model.snmp.proxy.impl.SnmpIndexColumn;
import org.snmp4j.smi.MaxAccess;
import org.snmp4j.smi.OID;
import org.snmp4j.smi.SubIndexInfo;

import java.util.ArrayList;
import java.util.List;

/**
 * The <code>TableHelper</code> class provides helper methods to format table titles and data.
 *
 * @author Frank Fock
 */
public class TableHelper {

  protected static SnmpProxyColumnFactory<String> columnFactory = new StringSnmpProxyColumnFactory();

  /**
   * Return column titles for a list of object names and a common prefix of the names which should not be included
   * in the titles.
   * @param prefix
   *    the prefix to remove from the object names to form the returned column title.
   * @param names
   *    a list of object names that should be used for the titles.
   * @return
   *    a list of the same size than <code>names</code> but each title without the specified <code>prefix</code>
   *    if present.
   */
  public static List<String> getTitlesFromNames(String prefix, List<String> names) {
    ArrayList<String> v = new ArrayList<String>(names.size() + 1);
    String commonPrefix = trimTitleSuffix(prefix);
    boolean match = false;
    while ((!match) && (commonPrefix.length() > 0)) {
      match = true;
      for (String s : names) {
        if (!s.startsWith(commonPrefix)) {
          commonPrefix = trimTitleSuffix(commonPrefix);
          match = false;
          break;
        }
      }
    }
    if (match) {
      for (String s : names) {
        v.add(s.substring(commonPrefix.length()));
      }
      return v;
    }
    else {
      return names;
    }
  }

  /**
   * Removes the last name suffix from the given title.
   * @param title
   *    a camel case title.
   * @return
   *    <code>title</code> with the last camel case word removed.
   */
  protected static String trimTitleSuffix(String title) {
    String s = title;
    while ((s.length() > 2) &&
        (Character.isLowerCase(s.charAt(s.length() - 1)))) {
      s = s.substring(0, s.length() - 1);
    }
    if (s.length() < 2) {
      return "";
    }
    return s.substring(0, s.length() - 1);
  }

  /**
   * Create the column proxy elements based on a table entry object type.
   * @param child
   *
   * @return
   */
  public static List<SnmpProxyColumn<String>> buildColumns(SmiObjectType child) {
    return buildColumns(child, String.class, columnFactory);
  }

  public static <T> List<SnmpProxyColumn<T>> buildColumns(SmiObjectType child, Class<T> valueClass,
                                                          SnmpProxyColumnFactory<T> columnFactory) {
    List<SmiObjectType> smiObjectTypes = child.getColumns();
    return getSnmpProxyColumns(child, valueClass, columnFactory, smiObjectTypes);
  }

  /**
   * Get the proxy columns for the specified SNMP table definition.
   * @param parentObjectType
   *   the parent (i.e., table entry) object type.
   * @param valueClass
   *   the value class of T.
   * @param columnFactory
   *   The factory to be used to create the column proxies.
   * @param smiObjectTypes
   *   The columnar object types. If a columnar object type equals <code>parentObjectType</code>, the column title
   *   will be the parent's full object name. Otherwise, the common prefix of the column names and the parent name
   *   will be removed from the column titles.
   * @param <T>
   *   The value type of the proxies.
   * @return
   *   a list of column snmp proxies corresponding to the number and position of the provided smiObjectTypes.
   */
  public static <T> List<SnmpProxyColumn<T>> getSnmpProxyColumns(SmiObjectType parentObjectType, Class<T> valueClass,
                                                                  SnmpProxyColumnFactory<T> columnFactory,
                                                                  List<SmiObjectType> smiObjectTypes) {
    ArrayList<SnmpProxyColumn<T>> columns = new ArrayList<SnmpProxyColumn<T>>(smiObjectTypes.size());
    List<String> columnNames = new ArrayList<String>(smiObjectTypes.size());
    for (SmiObject col : smiObjectTypes) {
      columnNames.add(col.getObjectName());
    }
    List<String> headers = getTitlesFromNames(parentObjectType.getObjectName(), columnNames);
    for (int i=0; i< smiObjectTypes.size(); i++) {
      SmiObjectType objectType = smiObjectTypes.get(i);
      columns.add(columnFactory.createSnmpProxyColumn(objectType.getOID(), valueClass, objectType.getMaxAccess(),
          objectType.getSyntax().getSyntax(),
          objectType.getMinValueLength(), objectType.getMaxValueLength(),
          parentObjectType.equals(objectType) ? parentObjectType.getObjectName() : headers.get(i)));
    }
    return columns;
  }

  public static List<SnmpProxyIndexColumn<String>> buildIndexColumns(SmiObjectType child) {
    return buildIndexColumns(child, String.class, columnFactory);
  }

  public static <T> List<SnmpProxyIndexColumn<T>> buildIndexColumns(SmiObjectType child, Class<T> valueClass,
                                                                    SnmpProxyColumnFactory<T> columnFactory) {
    List<SmiObjectTypeIndex> smiObjectTypes = child.getIndexObjects();
    ArrayList<SnmpProxyIndexColumn<T>> indexColumns = new ArrayList<SnmpProxyIndexColumn<T>>(smiObjectTypes.size());
    List<String> columnNames = new ArrayList<String>(smiObjectTypes.size());
    for (SmiObject col : smiObjectTypes) {
      columnNames.add(col.getObjectName());
    }
    List<String> headers = getTitlesFromNames(child.getObjectName(), columnNames);
    for (int i=0; i< smiObjectTypes.size(); i++) {
      SmiObjectTypeIndex objectType = smiObjectTypes.get(i);
      indexColumns.add(columnFactory.createSnmpProxyIndexColumn(objectType.getOID(), valueClass,
          objectType.getSyntax().getSyntax(),
          objectType.getMinValueLength(), objectType.getMaxValueLength(), headers.get(i), objectType.getIndexInfo()));
    }
    return indexColumns;
  }

  public static class StringSnmpProxyColumnFactory implements SnmpProxyColumnFactory<String>  {

    @Override
    public SnmpProxyColumn<String> createSnmpProxyColumn(OID oid, Class<String> valueClass,
                                                         MaxAccess smiMaxAccess, int smiSyntax,
                                                         int minWidth, int maxWidth, String header) {
      return new SnmpColumn<String>(oid, valueClass, smiMaxAccess, smiSyntax, minWidth, maxWidth, header);
    }

    @Override
    public SnmpIndexColumn<String> createSnmpProxyIndexColumn(OID oid, Class<String> valueClass,
                                                              int smiSyntax,
                                                              int minWidth, int maxWidth, String header,
                                                              SubIndexInfo smiIndexInfo) {
      return new SnmpIndexColumn<String>(oid, valueClass, smiSyntax, minWidth, maxWidth, header, smiIndexInfo);
    }
  }
}
