/*_############################################################################
  _## 
  _##  SNMP4J-Model - ColumnList.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.proxy.impl;

import org.snmp4j.model.snmp.proxy.ChangeSet;
import org.snmp4j.model.snmp.proxy.SnmpProxyRow;
import org.snmp4j.model.snmp.proxy.SnmpTable;
import org.snmp4j.smi.OID;

import java.util.ArrayList;
import java.util.Collection;

/**
 * The <code>ColumnList</code> implements a virtual column set based on a table and a row index.
 * The ColumnList provides write access to the columns of the row identified by the index given at
 * creation time of the given table.
 *
 * @param <C>
 *   specifies the column implementation class which has to extend {@link SnmpColumn}.
 * @param <IC>
 *   specifies the index column implementation class which has to extend {@link SnmpIndexColumn}.
 * @param <IT>
 *   specifies the
 * @author Frank Fock
 */
public class ColumnList<C extends SnmpColumn,IC extends SnmpIndexColumn, IT, T, R extends SnmpProxyRow<IT,T>>
    extends ArrayList<T> {

  private OID rowIndex;
  private SnmpTable<C,IC,IT,T,R> table;

  public ColumnList(Collection<? extends T> c, OID rowIndex, SnmpTable<C,IC,IT,T,R> table) {
    super(c);
    this.rowIndex = rowIndex;
    this.table = table;
  }

  @Override
  public T set(int columnIndex, T value) {
    if (value != null && value != super.get(columnIndex) && !value.equals(super.get(columnIndex))) {
      //System.out.println("Setting col " + columnIndex + " to " + value + " of row " + rowIndex );
      if (columnIndex < table.getColumnCount()) {
        if ((columnIndex >= table.getIndexColumns().size()) && (columnIndex < size())) {
          C column = table.getColumns().get(columnIndex);
          if (column.getSmiMaxAccess().isWritable()) {
            OID instanceOID = new OID(column.getOID().getValue(), rowIndex.getValue());
            SnmpValuesChangeSet changeSet = new SnmpValuesChangeSet();
            changeSet.addChange(new SnmpValueChange(instanceOID,
                column.getValueClass(), value, get(columnIndex)));
            SnmpValuesChangeSet validatedChangeSet = table.getCommitter().validateChange(changeSet);
            if (validatedChangeSet.getStatus() == ChangeSet.ChangeStatus.PREPARED) {
              table.getCommitter().addChange(table, changeSet);
              value = (T) validatedChangeSet.getChanges().get(0).getNewValue();
            }
            table.fireTableCellUpdated(table.getRowNumber(rowIndex), columnIndex);
          }
          return super.set(columnIndex, value);
        }
      }
    }
    return get(columnIndex);
  }
}
