/*_############################################################################
  _## 
  _##  SNMP4J-Model - SnmpScalarImplRO.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.proxy.impl;

import org.snmp4j.Target;
import org.snmp4j.model.mapper.VariableBindingMapper;
import org.snmp4j.model.snmp.spi.SnmpProxyReloadListener;
import org.snmp4j.model.snmp.proxy.SnmpScalarRO;
import org.snmp4j.model.snmp.proxy.TaskId;
import org.snmp4j.model.snmp.spi.SnmpException;
import org.snmp4j.model.snmp.spi.SnmpService;
import org.snmp4j.smi.OID;
import org.snmp4j.smi.OctetString;
import org.snmp4j.smi.VariableBinding;

import java.util.Collections;
import java.util.List;

/**
 * The <code>SnmpScalarImpl</code> class implements a SNMP scalar proxy with read-only access.
 * @param <T>
 *   specifies the model value type of the scalar proxy.
 *
 * @author Frank Fock
 */
public class SnmpScalarImplRO<T> extends SnmpProxyObjectImpl implements SnmpScalarRO<T> {

  private T value;
  private Class<T> valueClass;
  private int snmpSyntax;

  /**
   * Creates a {@link org.snmp4j.model.snmp.proxy.SnmpProxyObject} for a scalar SNMP OBJECT-TYPE.
   * @param oid
   *    the OID of the scalar instance value (thus including the .0 suffix).
   * @param context
   *    the SNMP context associated with this proxy object.
   * @param value
   *    the mode value of the scalar.
   * @param valueClass
   *    the model value class.
   * @param snmpSyntax
   *    the SNMP syntax (BER syntax code) of the OBJECT-TYPE's base syntax.
   */
  public SnmpScalarImplRO(OID oid, OctetString context, T value, Class<T> valueClass, int snmpSyntax) {
    super(context, oid);
    this.value = value;
    this.valueClass = valueClass;
    this.snmpSyntax = snmpSyntax;
  }

  @Override
  public TaskId reload(SnmpService snmpService, Target target, VariableBindingMapper variableBindingMapper,
                       SnmpProxyReloadListener reloadListener) {
    try {
      OID instanceOID = new OID(oid.getValue(), 0);
      List<? extends VariableBinding> vbs = snmpService.get(target, Collections.singletonList(instanceOID));
      setValue((T)variableBindingMapper.toObject(vbs.get(0), getValueClass()));
      if (reloadListener != null) {
        reloadListener.reloadFinished(this);
      }
    } catch (SnmpException e) {
      if (reloadListener != null) {
        // TODO: logging
        reloadListener.reloadFailed(this, e.getErrorStatus());
      }
    }
    return null;
  }

  @Override
  public boolean waitForTask(TaskId taskId, long timeout) {
    return false;
  }

  @Override
  public String getName() {
    return getOID().toString();
  }

  @Override
  public T getValue() {
    return value;
  }

  /**
   * Sets the model value of this scalar proxy without changing the SNMP value of the target SNMP entity.
   * @param value
   *   the new model value.
   */
  public void setValue(T value) {
    this.value = value;
  }

  @Override
  public Class<T> getValueClass() {
    return valueClass;
  }

  @Override
  public int getSnmpSyntax() {
    return snmpSyntax;
  }

  @Override
  public String toString() {
    return "SnmpScalarImplRO{" +  super.toString()+
        ", snmpSyntax=" + snmpSyntax +
        ", value=" + value +
        ", valueClass=" + valueClass +
        "} ";
  }
}
