/*_############################################################################
  _## 
  _##  SNMP4J-Model - SnmpTableImpl.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.proxy.impl;

import org.snmp4j.Target;
import org.snmp4j.model.mapper.VariableBindingMapper;
import org.snmp4j.model.snmp.proxy.*;
import org.snmp4j.model.snmp.spi.*;
import org.snmp4j.smi.OID;
import org.snmp4j.smi.OctetString;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * The SnmpTableImpl class implements a writable SNMP table that implements the Java
 * {@link javax.swing.table.TableModel}.
 *
 * @param <C>
 *   the proxy column type fo the table proxy.
 * @param <IC>
 *   the index proxy column type for the table proxy.
 * @param <IT>
 *   the model value type for index columns.
 * @param <T>
 *   the model value type for the columnar model values.
 * @param <R>
 *   the proxy row type for the rows of the table.
 *
 * @author Frank Fock
 */
public class SnmpTableImpl<IC extends SnmpProxyIndexColumn, C extends SnmpProxyColumn,IT, T, R extends SnmpProxyRow<IT,T>>
    extends SnmpTableImplRO<C, IC, IT, T, R> implements SnmpTable<C, IC, IT, T, R> {

  private SnmpProxyCommitter committer;
  private SnmpCellUpdateListener cellUpdateListener;

  /**
   * Creates a proxy table implementation based on the SNMP table entry object identifier and lists with the
   * index and regular column proxy objects as well as a page size to support paged table models.
   * @param tableEntryOID
   *    the OID of the SNMP table Entry object (thus the table OID inclusive the entry suffix <code>.1</code>).
   * @param context
   *    the SNMP context associated with this proxy object.
   * @param snmpProxyCommitter
   *    the proxy committer used to commit changes to model values to the associated target SNMP entity.
   * @param rowFactory
   *   the factory for new model rows.
   * @param indexColumns
   *   the index proxy column objects.
   * @param columns
   *   the proxy column objects.
   * @param pageSize
   *   the page size for paged table models. A zero (or less) size disables paging.
   */
  public SnmpTableImpl(OID tableEntryOID, OctetString context, SnmpProxyCommitter snmpProxyCommitter,
                       SnmpProxyRowFactory<R, IT, T> rowFactory,
                       List<IC> indexColumns, List<C> columns, int pageSize) {
    super(tableEntryOID, context, rowFactory, columns, indexColumns, pageSize);
    this.committer = snmpProxyCommitter;
  }

  @Override
  public SnmpProxyCommitter getCommitter() {
    return committer;
  }

  @Override
  public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
    if ((rowIndex >= 0) && (columnIndex < getColumnCount()) && (rowIndex < rows.size())) {
      R row = rows.get(rowIndex);
      if ((columnIndex >= indexColumns.size()) && (columnIndex - indexColumns.size() < row.size())) {
        C column = columns.get(columnIndex - indexColumns.size());
        OID instanceOID = new OID(column.getOID().getValue(), row.getIndex().getValue());
        SnmpValueChange change = new SnmpValueChange(instanceOID,
            column.getValueClass(), aValue, row.getValue(columnIndex - indexColumns.size()));
        SnmpValuesChangeSet changeSet = new SnmpValuesChangeSet();
        changeSet.addChange(change);
        SnmpValuesChangeSet validatedChangeSet = committer.validateChange(changeSet);
        if (validatedChangeSet.getStatus() == ChangeSet.ChangeStatus.PREPARED) {
          committer.addChange(this, validatedChangeSet);
          row.setValue(columnIndex - indexColumns.size(), (T)validatedChangeSet.getChanges().get(0).getNewValue());
          cellUpdated(rowIndex, columnIndex, validatedChangeSet.getChanges().get(0));
        }
        else {
          cellUpdateValidationFailure(rowIndex, columnIndex, validatedChangeSet.getChanges().get(0));
          row.setValue(columnIndex - indexColumns.size(), (T)aValue);
        }
        fireTableCellUpdated(rowIndex, columnIndex);
      }
    }
    // @TODO logging
  }

  public SnmpCellUpdateListener getCellUpdateListener() {
    return cellUpdateListener;
  }

  public void setCellUpdateListener(SnmpCellUpdateListener cellUpdateListener) {
    this.cellUpdateListener = cellUpdateListener;
  }

  protected void cellUpdateValidationFailure(int row, int column, SnmpValueChange change) {
    if (cellUpdateListener != null) {
      cellUpdateListener.cellUpdateFailed(row, column, change);
    }
  }

  protected void cellUpdated(int row, int column, SnmpValueChange change) {
    if (cellUpdateListener != null) {
      cellUpdateListener.cellUpdated(row, column, change);
    }
  }

  @Override
  public boolean isCellEditable(int rowIndex, int columnIndex) {
    return columnIndex > indexColumns.size() &&
        columns.get(columnIndex - indexColumns.size()).getSmiMaxAccess().isWritable();
  }

  public OID getCellOID(int column, OID index) {
    return new OID(getColumns().get(column).getOID().getValue(), index.getValue());
  }

  @Override
  public void insertRow(R newRow) {
    OID newIndex = newRow.getIndex();
    synchronized (rows) {
      int insertIndex = Collections.binarySearch(rows, newIndex, new SnmpProxyRowOidComparator());
      if (insertIndex < 0) {
        insertIndex = Math.abs(insertIndex)-1;
      }
      rows.add(insertIndex, newRow);
      fireTableRowsInserted(insertIndex, insertIndex);
      SnmpValuesChangeSet change = new SnmpValuesChangeSet();
      for (int j=0; j < newRow.size(); j++) {
        change.addChange(new SnmpValueChange(getCellOID(j, newIndex), (Class)getColumnClass(j),
            newRow.getValue(j), null));
      }
      committer.addChange(this, change);
    }
  }

  /**
   * Removes a row from the model without changing the associated SNMP table.
   * @param rowIndex
   *    the zero based row index.
   * @return
   *    the removed row.
   */
  public R removeRow(int rowIndex) {
    return rows.remove(rowIndex);
  }

  /**
   * Removes a row from the model without changing the associated SNMP table.
   * @param rowIndex
   *    the SNMP row index (i.e., the OID suffix that identifies the row in the table).
   * @return
   *    the removed row or <code>null</code> if such a row could not be found.
   */
  public R removeRow(OID rowIndex) {
    int i = getRowNumber(rowIndex);
    if (i >= 0 && i < rows.size()) {
      return removeRow(i);
    }
    return null;
  }

  @Override
  public void clear() {
    rows.clear();
  }

  @Override
  public void commitSuccess(List<SnmpValuesChangeSet> committedChanges) {

  }

  @Override
  public void commitFailure(List<SnmpValuesChangeSet> failedChanges, SnmpErrorStatus errorStatus, int errorIndex) {

  }

  @Override
  public TaskId reload(SnmpService snmpService, Target target, VariableBindingMapper variableBindingMapper,
                       SnmpProxyReloadListener reloadListener) {
    TaskId taskId = super.reload(snmpService, target, variableBindingMapper, reloadListener);
    committer.removeAllChanges(this);
    return taskId;
  }

  /**
   * Gets the row number for the specified SNMP row index value.
   * @param rowIndex
   *    a SNMP row index (i.e., the OID suffix that identifies a row in the target table).
   * @return
   *    the row number the is associated with the given SNMP row index or a negative number whose abstract value
   *    identifies where such a row could be inserted in the row model.
   */
  public int getRowNumber(OID rowIndex) {
    return Collections.binarySearch(rows, rowIndex, new SnmpProxyRowOidComparator());
  }

  @Override
  public boolean doChange(SnmpValuesChangeSet changeSet) {
    return doChange(changeSet, true);
  }

  /**
   * Performs a change by validating it using the {@link SnmpProxyCommitter}. Optionally the change is then also
   * added the the proxy committer queue.
   * @param changeSet
   *   a {@link SnmpValuesChangeSet} that contains changes to model objects that refer to cells of this proxy table.
   * @param addToCommitter
   *   if <code>true</code>, the changes will also be added to the proxy committer change queue for committing the
   *   changes to the target SNMP table.
   * @return
   *   <code>true</code> if the change set was valid. If not, no changes to the model and the target SNMP table
   *   are performed.
   */
  public boolean doChange(SnmpValuesChangeSet changeSet, boolean addToCommitter) {
    SnmpValuesChangeSet validatedChangeSet = committer.validateChange(changeSet);
    if (validatedChangeSet.getStatus() == ChangeSet.ChangeStatus.FAILURE) {
      return false;
    }
    if (addToCommitter) {
      committer.addChange(this, validatedChangeSet);
    }
    for (SnmpValueChange change : changeSet.getChanges()) {
      OID oid = change.getOid();
      SnmpCell cell = getCell(oid);
      int r = cell.getRow();
      if (r >= 0) {
        R row = rows.get(r);
        int columnIndex = cell.getColumn();
        //System.out.println("doChange: columnIndex="+columnIndex);
        for (int i=0; i<validatedChangeSet.getChanges().size(); i++) {
          row.setValue(columnIndex, (T)validatedChangeSet.getChanges().get(i).getNewValue());
        }
        fireTableCellUpdated(r, columnIndex+indexColumns.size());
      }
    }
    return true;
  }

  @Override
  public boolean undoChange(SnmpValuesChangeSet changeSet) {
    SnmpValuesChangeSet undoChangeSet =
        new SnmpValuesChangeSet(new ArrayList<SnmpValueChange>(), changeSet.getUserReference());
    for (SnmpValueChange change : changeSet.getChanges()) {
      undoChangeSet.addChange(new SnmpValueChange(change.getOid(), change.getValueClass(),
          change.getOldValue(), change.getNewValue()));
    }
    return doChange(undoChangeSet, ChangeSet.ChangeStatus.COMMITTED.equals(changeSet.getStatus()));
  }
}
