/*_############################################################################
  _## 
  _##  SNMP4J-Model - SnmpTableImplRO.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.proxy.impl;

import org.snmp4j.Target;
import org.snmp4j.model.mapper.VariableBindingMapper;
import org.snmp4j.model.snmp.proxy.*;
import org.snmp4j.model.snmp.spi.*;
import org.snmp4j.smi.*;

import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableColumnModel;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;
import java.util.*;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

/**
 * The {@link SnmpTableImplRO} class implements a proxy for a read-only SNMP table. Although it can be used
 * on a read-writable SNMP table too, modifications of the proxy table model cannot be committed to the
 * backing agent.
 *
 * @author Frank Fock
 */
public class SnmpTableImplRO<C extends SnmpProxyColumn, IC extends SnmpProxyIndexColumn, IT, T, R extends SnmpProxyRow<IT,T>>
    extends AbstractTableModel implements SnmpTableRO<IC, C, IT, T, R>, Collection<R>
{

  private static final long serialVersionUID = -8026396218277991810L;

  protected List<C> columns = new ArrayList<C>();
  protected List<IC> indexColumns = new ArrayList<IC>();
  protected List<R> rows = Collections.synchronizedList(new ArrayList<R>());
  protected List<R> updatingRows;
  protected OID tableEntryOID;
  protected int pageSize;
  protected int currentIndex;
  protected Object[] minIndex;
  protected SnmpProxyRowFilter rowFilter;
  protected SnmpProxyRowFactory<R,IT,T> rowFactory;
  protected ConcurrentHashMap<OID, RowUpdateError> rowErrors = new ConcurrentHashMap<OID, RowUpdateError>();
  protected OctetString context;

  /**
   * Creates a proxy table implementation based on the SNMP table entry object identifier and lists with the
   * index and regular column proxy objects as well as a page size to support paged table models.
   * @param tableEntryOID
   *    the OID of the SNMP table Entry object (thus the table OID inclusive the entry suffix <code>.1</code>).
   * @param context
   *    the SNMP context associated with this proxy object.
   * @param rowFactory
   *   the factory for new model rows.
   * @param indexColumns
   *   the index proxy column objects.
   * @param columns
   *   the proxy column objects.
   * @param pageSize
   *   the page size for paged table models. A zero (or less) size disables paging.
   */
  public SnmpTableImplRO(OID tableEntryOID, OctetString context,
                         SnmpProxyRowFactory<R,IT,T> rowFactory, List<C> columns, List<IC> indexColumns, int pageSize) {
    this.tableEntryOID = tableEntryOID;
    this.context = context;
    this.rowFactory = rowFactory;
    this.columns = columns;
    this.indexColumns = indexColumns;
    this.pageSize = pageSize;
  }

  @Override
  public OctetString getContext() {
    return context;
  }

  @Override
  public String getColumnName(int column) {
    if (column < indexColumns.size()) {
      return indexColumns.get(column).toString();
    }
    else if (column - indexColumns.size() < columns.size()) {
      return columns.get(column - indexColumns.size()).toString();
    }
    return null;
  }

  @Override
  public int findColumn(String columnName) {
    return super.findColumn(columnName);
  }

  @Override
  public Class<?> getColumnClass(int columnIndex) {
    return super.getColumnClass(columnIndex);
  }

  @Override
  public OID getOID() {
    return tableEntryOID;
  }

  @Override
  public SnmpCell getCell(OID instanceOID) {
    OID columnOID;
    OID rowIndex;
    if (instanceOID.size() == tableEntryOID.size() + 1) {
      columnOID = tableEntryOID;
      rowIndex = new OID(new int[] { 0 });
    }
    else {
      columnOID = new OID(instanceOID.getValue(), 0, tableEntryOID.size() + 1);
      rowIndex = new OID(instanceOID.getValue(), tableEntryOID.size() + 1, instanceOID.size() - columnOID.size());
    }
    int row = getRowNumber(rowIndex);
    int column = getColumnNumber(columnOID);
    return new SnmpCell(row, column, instanceOID, rowIndex);
  }

  @Override
  public TaskId reload(SnmpService snmpService, Target target, final VariableBindingMapper variableBindingMapper,
                       final SnmpProxyReloadListener reloadListener) {
    final List<C> columns = getColumns();
    List<OID> colOIDs = new ArrayList<OID>(columns.size());
    for (C column : columns) {
      colOIDs.add(column.getOID());
    }
    final List<IC> indexColumns = getIndexColumns();
    OID minIndexOID = null;
    OID maxIndexOID = null;
    try {
      SnmpRowListener rowListener = new SnmpRowListener() {
        private volatile boolean finished;
        @Override
        public void beginUpdate(Object userObject) {
          SnmpTableImplRO.this.beginUpdate(userObject);
        }

        @Override
        @SuppressWarnings("unchecked")
        public boolean nextRow(SnmpRow nextRow, Object userObject) {
          List<VariableBinding> indexVBs = getIndexVariableBindings(nextRow.getIndex());
          updateRow(userObject, getRowFactory().createProxyRow(nextRow.getIndex(),
              (List<IT>)SnmpProxyObjectHelper.mapVBsToObjects(variableBindingMapper, indexVBs, indexColumns),
              (List<T>)SnmpProxyObjectHelper.mapVBsToObjects(variableBindingMapper, nextRow.getVariableBindings(), columns)));
          return true;
        }

        @Override
        public synchronized void endUpdate(SnmpErrorStatus errorStatus, Object userObject) {
          finishUpdate(userObject);
          if (reloadListener != null) {
            reloadListener.reloadFinished(SnmpTableImplRO.this);
          }
          finished = true;
          this.notify();
        }

        @Override
        public boolean isFinished() {
          return finished;
        }
      };
      snmpService.getTable(target, colOIDs, minIndexOID, maxIndexOID, rowListener, rowListener);
      return rowListener;
    } catch (SnmpException e) {
      if (reloadListener != null) {
        reloadListener.reloadFailed(this, e.getErrorStatus());
      }
      e.printStackTrace();
    }
    return null;
  }

  @Override
  public boolean waitForTask(TaskId taskId, long timeout) {
    try {
      synchronized (taskId) {
        taskId.wait(timeout);
      }
    } catch (InterruptedException e) {
    }
    return taskId.isFinished();
  }

  @Override
  public List<C> getColumns() {
    return columns;
  }

  @Override
  public List<IC> getIndexColumns() {
    return indexColumns;
  }

  @Override
  public void setPageSize(int numRows) {
     this.pageSize = numRows;
  }

  @Override
  public int getPageSize() {
    return pageSize;
  }

  @Override
  public void setCurrentIndex(int startPos) {
    this.currentIndex = startPos;
  }

  @Override
  public int getCurrentIndex() {
    return currentIndex;
  }

  @Override
  public void setRowFilter(SnmpProxyRowFilter rowFilter) {
    this.rowFilter = rowFilter;
  }

  @Override
  public SnmpProxyRowFilter getRowFilter() {
    return rowFilter;
  }

  public Object[] getMinIndex() {
    return minIndex;
  }

  @Override
  public List<VariableBinding> getIndexVariableBindings(OID index) {
    List<OID> indexOIDs = getIndexOIDs(index, getIndexColumns());
    return getIndexVariables(getIndexColumns(), indexOIDs);
  }

  public void setMinIndex(Object[] minIndex) {
    this.minIndex = minIndex;
  }

  @Override
  public SnmpProxyRowFactory<R,IT,T> getRowFactory() {
    return rowFactory;
  }

  public void setRowFactory(SnmpProxyRowFactory<R,IT,T> rowFactory) {
    this.rowFactory = rowFactory;
  }

  @Override
  public void beginUpdate(Object userObject) {
    updatingRows = new ArrayList<R>();
  }

  @Override
  public void updateRow(Object userObject, R updateRow) {
    if (rowFilter == null || rowFilter.passesFilter(updateRow)) {
      updatingRows.add(updateRow);
    }
  }

  @Override
  public synchronized void finishUpdate(Object userObject) {
    if (updatingRows != null) {
      rows = updatingRows;
      fireTableStructureChanged();
    }
  }

  @Override
  public synchronized int getRowCount() {
    return Math.min(rows.size(), pageSize);
  }

  @Override
  public int getColumnCount() {
    return indexColumns.size() + columns.size();
  }

  @Override
  public synchronized Object getValueAt(int rowIndex, int columnIndex) {
    R row = rows.get(getNonPagedRowIndex(rowIndex));
    if (columnIndex < indexColumns.size()) {
      return row.getIndexObjects().get(columnIndex);
    }
    return row.getValue(columnIndex - indexColumns.size());
  }

  protected int getNonPagedRowIndex(int pageRowIndex) {
    return (pageRowIndex % pageSize) + (currentIndex * pageSize);
  }

  public static VariableBinding getIndexVariable(SnmpProxyIndexColumn indexColumn, OID subIndex) {
    try {
      Variable variable;
      switch (indexColumn.getSnmpSyntax()) {
        case SMIConstants.SYNTAX_OCTET_STRING:
          if ((indexColumn.hasImpliedLength()) || (indexColumn.getMinLength() == indexColumn.getMaxLength())) {
            variable = new OctetString(subIndex.toByteArray());
          }
          else {
            variable = new OctetString(subIndex.toByteArray(), 1, subIndex.size()-1);
          }
          break;
        case SMIConstants.SYNTAX_UNSIGNED_INTEGER32:
          variable = new UnsignedInteger32((subIndex.get(subIndex.size() - 1)));
          break;
        case SMIConstants.SYNTAX_INTEGER32:
          variable = new Integer32((subIndex.get(subIndex.size() - 1)));
          break;
        case SMIConstants.SYNTAX_TIMETICKS:
          variable = new TimeTicks((subIndex.get(subIndex.size() - 1)));
          break;
        case SMIConstants.SYNTAX_IPADDRESS:
          variable = new IpAddress(subIndex.toString());
          break;
        default:
          if ((indexColumn.hasImpliedLength()) || (indexColumn.getMinLength() == indexColumn.getMaxLength())) {
            variable = subIndex;
          }
          else {
            variable = new OID(subIndex.getValue(), 1, subIndex.size()-1);
          }
          break;
      }
      return new VariableBinding(indexColumn.getOID(), variable);
    }
    catch (Exception e) {
      return new VariableBinding(indexColumn.getOID(), subIndex);
    }
  }

  public static List<VariableBinding> getIndexVariables(List<? extends SnmpProxyIndexColumn> indexes,
                                                        List<? extends OID> indexValues) {
    int sz = Math.min(indexes.size(), indexValues.size());
    ArrayList<VariableBinding> vbs = new ArrayList<VariableBinding>(sz);
    for (int i=0; i<sz; i++) {
      vbs.add(getIndexVariable(indexes.get(i), indexValues.get(i)));
    }
    return vbs;
  }

  public static List<VariableBinding> getIndexVariableTemplates(List<? extends SnmpProxyIndexColumn> indexes) {
    int sz = indexes.size();
    ArrayList<VariableBinding> vbs = new ArrayList<VariableBinding>(sz);
    for (SnmpProxyIndexColumn indexColumn : indexes) {
      Variable v = AbstractVariable.createFromSyntax(indexColumn.getSmiSyntax());
      OID oid = indexColumn.getOID();
      vbs.add(new VariableBinding(oid, v));
    }
    return vbs;
  }

  /**
   * Split a table index into an array of object IDs each representing the
   * value of its corresponding index object. For example if a table's index
   * would be defined as INDEX { ifIndex, ipAddress } and the index given
   * would be "1.127.0.0.1" the resulting array would be { "1", "127.0.0.1" }
   *
   * @param index
   *    an OID denoting a table's index value.
   * @param indexColumns
   *    an array of index object descriptions.
   * @return
   *    a list of OID instances with the same size as indexColumns. If
   *    the given index is not a valid object ID null is returned.
   */
  public List<OID> getIndexOIDs(OID index, List<? extends SnmpProxyIndexColumn> indexColumns) {
    OID indexSuffix = index;
    ArrayList<OID> subIndexes = new ArrayList<OID>(indexColumns.size());
    for (SnmpProxyIndexColumn indexColumn : indexColumns) {
      OID next;
      if (indexColumn.hasImpliedLength()) {
        next = indexSuffix;
      }
      else if ((!indexColumn.hasImpliedLength()) && (indexColumn.getMinLength() != indexColumn.getMaxLength())) {
        next = new OID(indexSuffix.getValue(), 0, indexSuffix.get(0) + 1);
      }
      else {
        next = new OID(indexSuffix.getValue(), 0, indexColumn.getMaxLength());
      }
      subIndexes.add(next);
      indexSuffix = new OID(indexSuffix.getValue(), next.size(), indexSuffix.size() - next.size());
    }
    return subIndexes;
  }

  @Override
  public TableColumnModel getTableColumnModel(int maxCharacterPixelWidth, int maxColumnCharacters) {
    DefaultTableColumnModel tableColumnModel = new DefaultTableColumnModel();
    for (int i=0; i<indexColumns.size(); i++) {
      IC indexColumn = indexColumns.get(i);
      TableColumn tableColumn = new TableColumn(i,
          maxCharacterPixelWidth * Math.min(maxColumnCharacters, indexColumn.getMaxWidth()));
      tableColumn.setHeaderValue(indexColumn.getHeader());
      tableColumnModel.addColumn(tableColumn);
    }
    for (int i=0; i<columns.size(); i++) {
      C column = columns.get(i);
      TableColumn tableColumn = new TableColumn(i+indexColumns.size(),
          maxCharacterPixelWidth * Math.min(maxColumnCharacters, column.getMaxWidth()));
      tableColumn.setHeaderValue(column.getHeader());
      tableColumnModel.addColumn(tableColumn);
    }
    return tableColumnModel;
  }

  @Override
  public void addError(List<SnmpValuesChangeSet> valuesChangeSets, SnmpErrorStatus errorStatus, int errorIndex) {
    int columnIndex = 0;
    SnmpValuesChangeSet errorChangeSet = null;
    for (SnmpValuesChangeSet changeSet : valuesChangeSets) {
      int offset = changeSet.getChanges().size();
      if (columnIndex + offset < errorIndex) {
        columnIndex += offset;
      }
      else {
        // errorIndex is 1-based
        columnIndex = (errorIndex - columnIndex) - 1;
        errorChangeSet = changeSet;
        break;
      }
    }
    OID cellOID = errorChangeSet != null ? errorChangeSet.getVariableBindings().get(columnIndex).getOid() : null;
    if (cellOID != null) {
      int columnOidLength = tableEntryOID.getValue().length + 1;
      OID columnOID = new OID(cellOID.getValue(), 0, columnOidLength);
      OID rowIndex = new OID(cellOID.getValue(), columnOidLength, cellOID.size() - columnOidLength);
      RowUpdateError rowUpdateError =
          new RowUpdateError(errorStatus, Math.max(getColumnNumber(columnOID), 0), errorChangeSet);
      rowErrors.put(rowIndex, rowUpdateError);
    }
  }

  @Override
  public SnmpErrorStatus getError(OID rowIndex, int column) {
    System.out.println(rowErrors);
    System.out.println(rowIndex + " : " + column);
    RowUpdateError rowUpdateError = rowErrors.get(rowIndex);
    if ((rowUpdateError != null) && ((rowUpdateError.columnIndex == column) || (rowUpdateError.columnIndex < 0))) {
      System.out.println(rowUpdateError);
      return rowUpdateError.errorStatus;
    }
    return null;
  }

  @Override
  public void clearAllErrors() {
    rowErrors.clear();
  }

  @Override
  public SnmpErrorStatus clearError(OID rowIndex) {
    RowUpdateError rowUpdateError = rowErrors.remove(rowIndex);
    if (rowUpdateError != null) {
      return rowUpdateError.errorStatus;
    }
    return null;
  }

  @Override
  public int size() {
    return getRowCount();
  }

  @Override
  public boolean isEmpty() {
    return rows.isEmpty();
  }

  @Override
  public boolean contains(Object o) {
    return rows.contains(o);
  }

  @Override
  public Iterator<R> iterator() {
    return rows.iterator();
  }

  @Override
  public Object[] toArray() {
    return rows.toArray();
  }

  @Override
  public <T> T[] toArray(T[] a) {
    return rows.toArray(a);
  }

  @Override
  public boolean add(R r) {
    return false;
  }

  @Override
  public boolean remove(Object o) {
    return rows.remove(o);
  }

  @Override
  public boolean containsAll(Collection<?> c) {
    return rows.containsAll(c);
  }

  @Override
  public boolean addAll(Collection<? extends R> c) {
    return rows.addAll(c);
  }

  @Override
  public boolean removeAll(Collection<?> c) {
    return rows.removeAll(c);
  }

  @Override
  public boolean retainAll(Collection<?> c) {
    return rows.retainAll(c);
  }

  @Override
  public void clear() {
    rows.clear();
  }

  @Override
  public int getRowNumber(OID index) {
    for (int i=0; i<rows.size(); i++) {
      if (rows.get(i).getIndex().equals(index)) {
        return i;
      }
    }
    return -1;
  }

  @Override
  public int getColumnNumber(OID columnOID) {
    for (int i=0; i<columns.size(); i++) {
      OID colOID = columns.get(i).getOID();
      if ((colOID.last() == columnOID.last()) &&
          (colOID.equals(columnOID))) {
        return i;
      }
    }
    return -1;
  }

  @Override
  public List<OID> getRowIndexes() {
    ArrayList<OID> rowIndexes = new ArrayList<OID>(rows.size());
    for (R row : rows) {
      rowIndexes.add(row.getIndex());
    }
    return rowIndexes;
  }

  @Override
  public OID getRowIndex(int rowNumber) {
    return rows.get(rowNumber).getIndex();
  }

  @Override
  public String toString() {
    return "SnmpTableImplRO{" + super.toString() +
        ", columns=" + columns +
        ", indexColumns=" + indexColumns +
        ", rows=" + rows +
        ", updatingRows=" + updatingRows +
        ", tableEntryOID=" + tableEntryOID +
        ", pageSize=" + pageSize +
        ", currentIndex=" + currentIndex +
        ", minIndex=" + Arrays.toString(minIndex) +
        ", rowFilter=" + rowFilter +
        ", rowFactory=" + rowFactory +
        '}';
  }

  protected class SnmpProxyRowOidComparator implements Comparator {

    @Override
    public int compare(Object o1, Object o2) {
      if ((!(o1 instanceof OID) && !(o1 instanceof SnmpProxyRow)) ||
          (!(o2 instanceof OID) && !(o2 instanceof SnmpProxyRow))) {
        throw new ClassCastException();
      }
      OID oid1,oid2;
      if (o1 instanceof SnmpProxyRow) {
        oid1 = ((SnmpProxyRow)o1).getIndex();
      }
      else {
        oid1 = (OID)o1;
      }
      if (o2 instanceof SnmpProxyRow) {
        oid2 = ((SnmpProxyRow)o2).getIndex();
      }
      else {
        oid2 = (OID)o2;
      }
      return oid1.compareTo(oid2);
    }
  }

  protected class SnmpColumnOidComparator implements Comparator {

    @Override
    public int compare(Object o1, Object o2) {
      if ((!(o1 instanceof OID) && !(o1 instanceof SnmpColumn)) ||
          (!(o2 instanceof OID) && !(o2 instanceof SnmpColumn))) {
        throw new ClassCastException();
      }
      OID oid1,oid2;
      if (o1 instanceof SnmpColumn) {
        oid1 = ((SnmpColumn)o1).getOID();
      }
      else {
        oid1 = (OID)o1;
      }
      if (o2 instanceof SnmpColumn) {
        oid2 = ((SnmpColumn)o2).getOID();
      }
      else {
        oid2 = (OID)o2;
      }
      return oid1.compareTo(oid2);
    }
  }


  protected class RowUpdateError {
    private SnmpErrorStatus errorStatus;
    private int columnIndex;
    private SnmpValuesChangeSet errorChangeSet;


    public RowUpdateError(SnmpErrorStatus errorStatus, int columnIndex, SnmpValuesChangeSet errorChangeSet) {
      this.errorStatus = errorStatus;
      this.columnIndex = columnIndex;
      this.errorChangeSet = errorChangeSet;
    }

    @Override
    public String toString() {
      return "RowUpdateError{" +
          "errorStatus=" + errorStatus +
          ", columnIndex=" + columnIndex +
          ", errorChangeSet=" + errorChangeSet +
          '}';
    }
  }

}
