/*_############################################################################
  _## 
  _##  SNMP4J-Model - SnmpErrorStatus.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.spi;

import org.snmp4j.PDU;
import org.snmp4j.mp.SnmpConstants;

/**
 * The <code>SnmpErrorStatus</code> represents the values of the SNMP error status
 * values.
 *
 * @author Frank Fock
 */
public enum SnmpErrorStatus {


  /**
   * Operation timed out (undefined error).
   */
  timeout(SnmpConstants.SNMP_ERROR_TIMEOUT),

  /**
   * GETNEXT (GETBULK) return OID <= requested OID.
   */
  lexicographicOrder(SnmpConstants.SNMP_ERROR_LEXICOGRAPHIC_ORDER),

  /**
   * The command responder returned a REPORT PDU that cannot be (automatically) resolved by the command generator.
   */
  report(SnmpConstants.SNMP_ERROR_REPORT),

  /**
   * IO error.
   */
  ioError(SnmpConstants.SNMP_ERROR_IO),

  /**
   * Operation success (no error).
   */
  noError(SnmpConstants.SNMP_ERROR_SUCCESS),
  /**
   * PDU encoding is too big for the transport used.
   */
  tooBig(SnmpConstants.SNMP_ERROR_TOO_BIG),
  /**
   * No such variable binding name, see error index.
   */
  noSuchName(SnmpConstants.SNMP_ERROR_NO_SUCH_NAME),

  /**
   * Bad value in variable binding, see error index.
   */
  badValue(SnmpConstants.SNMP_ERROR_BAD_VALUE),

  /**
   * The variable binding is read-only, see error index.
   */
  readOnly(SnmpConstants.SNMP_ERROR_READ_ONLY),

  /**
   * An unspecific error caused by a variable binding, see error index.
   */
  genErr(SnmpConstants.SNMP_ERROR_GENERAL_ERROR),

  /**
   * The variable binding is not accessible by the current MIB view, see error
   * index.
   */
  noAccess(SnmpConstants.SNMP_ERROR_NO_ACCESS),

  /**
   * The variable binding's value has the wrong type, see error index.
   */
  wrongType(SnmpConstants.SNMP_ERROR_WRONG_TYPE),

  /**
   * The variable binding's value has the wrong length, see error index.
   */
  wrongLength(SnmpConstants.SNMP_ERROR_WRONG_LENGTH),

  /**
   * The variable binding's value has a value that could under no circumstances
   * be assigned, see error index.
   */
  wrongValue(SnmpConstants.SNMP_ERROR_WRONG_VALUE),

  /**
   * The variable binding's value has the wrong encoding, see error index.
   */
  wrongEncoding(SnmpConstants.SNMP_ERROR_WRONG_ENCODING),

  /**
   * The specified object does not exists and cannot be created,
   * see error index.
   */
  noCreation(SnmpConstants.SNMP_ERROR_NO_CREATION),

  /**
   * The variable binding's value is presently inconsistent with the current
   * state of the target object, see error index.
   */
  inconsistentValue(SnmpConstants.SNMP_ERROR_INCONSISTENT_VALUE),

  /**
   * The resource needed to assign a variable binding's value is presently
   * unavailable, see error index.
   */
  resourceUnavailable(SnmpConstants.SNMP_ERROR_RESOURCE_UNAVAILABLE),

  /**
   * Unable to commit a value, see error index.
   */
  commitFailed(SnmpConstants.SNMP_ERROR_COMMIT_FAILED),

  /**
   * Unable to undo a committed value, see error index.
   */
  undoFailed(SnmpConstants.SNMP_ERROR_UNDO_FAILED),

  /**
   * Unauthorized access, see error index.
   */
  authorizationError(SnmpConstants.SNMP_ERROR_AUTHORIZATION_ERROR),

  /**
   * The variable's value cannot be modified, see error index.
   */
  notWritable(SnmpConstants.SNMP_ERROR_NOT_WRITEABLE),

  /**
   * The specified object does not exists and presently it cannot be created,
   * see error index.
   */
  inconsistentName(SnmpConstants.SNMP_ERROR_INCONSISTENT_NAME);


  private int errorStatus;

  SnmpErrorStatus(int errorStatus) {
    this.errorStatus = errorStatus;
  }

  /**
   * Returns the SNMP error status as integer value as defined
   * by {@link SnmpConstants}.
   * @return
   *    the integer value of this error status.
   */
  public int getErrorStatus() {
    return errorStatus;
  }

  public String toString() {
    return PDU.toErrorStatusText(getErrorStatus());
  }

  /**
   * Gets the SnmpErrorStatus enum value based on the given SNMP error status
   * as defined by {@link org.snmp4j.mp.SnmpConstants}.
   * @param errorStatus
   *   a SNMP error status value.
   * @return
   *   the corresponding SnmpErrorStatus enum or null if such a mapping thus not exists
   *   for the provided integer value.
   */
  public static SnmpErrorStatus fromSnmpErrorStatus(int errorStatus) {
    for (SnmpErrorStatus snmpErrorStatus : values()) {
      if (snmpErrorStatus.getErrorStatus() == errorStatus) {
        return snmpErrorStatus;
      }
    }
    return null;
  }

}
