/*_############################################################################
  _## 
  _##  SNMP4J-Model - SnmpProxyCommitter.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.spi;

import org.snmp4j.model.snmp.proxy.SnmpCommitListener;
import org.snmp4j.model.snmp.proxy.SnmpProxyObject;
import org.snmp4j.model.snmp.proxy.TaskId;
import org.snmp4j.model.snmp.proxy.impl.SnmpValuesChangeSet;

import java.util.List;

/**
 * Implementations of the <code>SnmpProxyCommitter</code> interface commit model changes to the
 * associated SNMP entity. Changes are committed atomically if they are put within the same change set.
 * Changes are queued until {@link #commitChanges(SnmpCommitListener)} is being called. Then all
 * queued changes are processed until no changes are left in the queue as it was at the time when
 * {@link #commitChanges(SnmpCommitListener)} has been called.
 *
 * @author Frank Fock
 */
public interface SnmpProxyCommitter {

  /**
   * Validate the change set provided.
   * @param change
   *   a change set.
   * @return
   *   <code>true</code> if the change set is valid that is, if all model values in the change can be mapped
   *   (converted) to SNMP values.
   */
  SnmpValuesChangeSet validateChange(SnmpValuesChangeSet change);

  /**
   * Adds a change on the specified proxy object to the queue.
   * @param proxyObject
   *   the proxy object that changed.
   * @param change
   */
  void addChange(SnmpProxyObject proxyObject, SnmpValuesChangeSet change);

  /**
   * Remove the last change from the queue for the specified proxy object.
   * @param proxyObject
   *    the proxy object for which the change to remove was added with
   *    {@link #addChange(SnmpProxyObject, SnmpValuesChangeSet)}.
   * @return
   *    the removed change set or <code>null</code> if there is no such change.
   */
  SnmpValuesChangeSet removeLastChangeSet(SnmpProxyObject proxyObject);

  /**
   * Remove all changes for a given proxy object.
   * @param proxyObject
   *    the proxy object for which all changes should be removed from the queue.
   * @return
   *    the removed changes (could be an empty list).
   */
  List<SnmpValuesChangeSet> removeAllChanges(SnmpProxyObject proxyObject);

  /**
   * Commit the changes in the current queue asynchronously in the background.
   * Note: changes that are queued after this call returns will not be committed.
   * @param commitListener
   *   a callback listener for commit status events.
   * @return
   *   the task ID for this commit operation.
   */
  TaskId commitChanges(SnmpCommitListener commitListener);

  /**
   * Gets the total number of changes in the queue (independently of a particular proxy object).
   * @return
   *    the number of changes in the queue.
   */
  int getNumberOfPendingChanges();

}
