/*_############################################################################
  _## 
  _##  SNMP4J-Model - SimpleMibBrowser.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.demo.mibbrowser;

import com.snmp4j.smi.SmiManager;
import com.snmp4j.smi.SmiObject;
import com.snmp4j.smi.SmiObjectType;
import com.snmp4j.smi.util.MemRepositoryDriver;
import org.snmp4j.MessageDispatcherImpl;
import org.snmp4j.SNMP4JSettings;
import org.snmp4j.Target;
import org.snmp4j.TransportMapping;
import org.snmp4j.asn1.BER;
import org.snmp4j.log.*;
import org.snmp4j.model.mapper.VariableBindingMapper;
import org.snmp4j.model.mapper.VariableBindingMapperJava;
import org.snmp4j.model.snmp.api.LocalSnmpService;
import org.snmp4j.model.snmp.api.TableHelper;
import org.snmp4j.model.snmp.proxy.*;
import org.snmp4j.model.snmp.proxy.impl.*;
import org.snmp4j.model.snmp.spi.SnmpService;
import org.snmp4j.mp.MPv1;
import org.snmp4j.mp.MPv2c;
import org.snmp4j.mp.MPv3;
import org.snmp4j.security.*;
import org.snmp4j.smi.OID;
import org.snmp4j.smi.TcpAddress;
import org.snmp4j.smi.TlsAddress;
import org.snmp4j.transport.AbstractTransportMapping;
import org.snmp4j.transport.DefaultTcpTransportMapping;
import org.snmp4j.transport.DefaultUdpTransportMapping;
import org.snmp4j.transport.TLSTM;
import org.snmp4j.util.ArgumentParser;
import org.snmp4j.util.SnmpConfigurator;

import javax.swing.*;
import javax.swing.table.AbstractTableModel;
import javax.swing.tree.TreePath;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * The SimpleMibBrowser is a demo application for the SNMP4J Model API. The MIB Browser
 * is able to selectable walk sub-trees of a MIB as well as browse and edit MIB tables.
 * It depends on SNMP4J-SMI-PRO API.
 *
 * @author Frank Fock
 * @version 1.0
 */
public class SimpleMibBrowser extends JFrame implements SnmpProxyRowFactory<SnmpProxyRowImpl<String,String>, String, String> {

  // initialize Java logging
  static {
    LogFactory.setLogFactory(new ConsoleLogFactory());
    BER.setCheckSequenceLength(false);
    LogFactory.getLogFactory().getRootLogger().setLogLevel(LogLevel.ALL);
  }

  private static final String SNMP4J_SMI_PRO_LICENSE_OPTIONS =
      "+L[s] ";
  private static final String MIB_OPTIONS =
      "+M[s] +m[s] ";
  private static final String V3_OPTIONS =
      "+a[s<MD5|SHA>] +A[s] +bc[i{=0}] +e[x] +E[x] "+
          "+y[s<DES|3DES|AES128|AES192|AES256|AES192p|AES256p>] +Y[s] +u[s] +l[x] +n[s] ";
  private static final String SNMP_OPTIONS =
      "+c[s] +r[i{=1}] +t[i{=5000}] +v[s{=3}<1|2c|3>] +Ors[i{=65535}] ";
  private static final String BULK_OPTIONS =
      "-Cr[i{=10}] -Cn[i{=0}] ";

  private static final String ADDRESS_PARAMETER =
      "#address[s<((udp|tcp):)?.*[/[0-9]+]?>] ";

  private static final String ALL_OPTIONS = MIB_OPTIONS +
      V3_OPTIONS + SNMP_OPTIONS + BULK_OPTIONS + SNMP4J_SMI_PRO_LICENSE_OPTIONS;

  private SmiManager smiManager;
  private SnmpService snmpService;
  private VariableBindingMapper variableBindingMapper;
  private SnmpProxyAdapter snmpProxyAdapter;

  private SimpleMibTreeTable mibTreeTable;

  private TreePath popupPath = null;

  private JPopupMenu popupMenuNode = new JPopupMenu();
  private JMenuItem menuItemBrowse = new JMenuItem();
  private JMenuItem menuItemGet = new JMenuItem();
  private JMenuItem menuItemTable = new JMenuItem();
  private Target target;

  public SimpleMibBrowser(Map<String,List<?>> commandLineParameters) throws IOException {
    smiManager = configureSnmp4jSMI(commandLineParameters);
    SnmpConfigurator snmpConfig = new SnmpConfigurator();
    target = snmpConfig.getTarget(commandLineParameters);

    AbstractTransportMapping transport;
    if (target.getAddress() instanceof TlsAddress) {
      transport = new TLSTM();
    }
    else if (target.getAddress() instanceof TcpAddress) {
      transport = new DefaultTcpTransportMapping();
    }
    else {
      transport = new DefaultUdpTransportMapping();
    }
    snmpService = new LocalSnmpService(new MessageDispatcherImpl(), Arrays.asList(new TransportMapping[] { transport}),
        Arrays.asList(new MPv1(), new MPv2c(), new MPv3()),
        SecurityModels.getCollection(new SecurityModel[]{ new USM(), new TSM() }),
        SecurityProtocols.getInstance().addDefaultProtocols());
    variableBindingMapper = new VariableBindingMapperJava();
    snmpProxyAdapter = new SnmpProxyAdapter(snmpService, target, variableBindingMapper);
    snmpConfig.configure(((LocalSnmpService)snmpService).getSnmp(), commandLineParameters);
    SNMP4JSettings.setNoGetBulk(true);
    initGUI();
  }

  private void initGUI() {
    setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);

    setLayout(new BorderLayout(0, 0));

    SimpleMibTreeTableModel treeTableModel = new SimpleMibTreeTableModel(createDataMibStructure());

    mibTreeTable = new SimpleMibTreeTable(treeTableModel);

    Container cPane = getContentPane();

    cPane.add(new JScrollPane(mibTreeTable), BorderLayout.CENTER);


    setSize(1000, 800);
    setLocationRelativeTo(null);

    initPopupMenu();
    mibTreeTable.addMouseListener(new MouseAdapter() {
      @Override
      public void mouseClicked(MouseEvent e) {
        //showPopupMenu(e);
        super.mouseClicked(e);
      }

      @Override
      public void mousePressed(MouseEvent e) {
        super.mousePressed(e);
        showPopupMenu(e);
      }

      @Override
      public void mouseReleased(MouseEvent e) {
        super.mouseReleased(e);
        showPopupMenu(e);
      }
    });
    try {
      snmpService.initTransports();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  private void showPopupMenu(MouseEvent e) {
    TreePath path = mibTreeTable.getTree().getPathForLocation(e.getX(), e.getY());
    if ((path != null) && ((e.isPopupTrigger()) ||
        (e.getModifiers() == InputEvent.META_MASK))) {
      // determine object type
      popupPath = path;
      SimpleMibDataNode node = (SimpleMibDataNode) path.getLastPathComponent();
      enablePopupMenu(node);
      popupMenuNode.setVisible(true);
      popupMenuNode.show(mibTreeTable, e.getX(), e.getY());
    }
  }

  private void enablePopupMenu(SimpleMibDataNode node) {

  }

  private void initPopupMenu() {
    //menuItemGet.setIcon(imagePlay16);
    menuItemGet.setMnemonic('G');
    menuItemGet.setToolTipText("Refresh value of the current OBJECT-TYPE (not applicable for tables");
    menuItemGet.setText("Get");
    menuItemGet.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(ActionEvent e) {
        doGetAction(e);
      }
    });
    menuItemTable.setMnemonic('T');
    menuItemTable.setToolTipText("Browse the current node as SNMP table");
    menuItemTable.setText("Table...");
    menuItemTable.addActionListener(new java.awt.event.ActionListener() {
      public void actionPerformed(ActionEvent e) {
        doTableAction(e);
      }
    });
    popupMenuNode.add(menuItemGet);
    popupMenuNode.add(menuItemTable);
  }

  private void doGetAction(ActionEvent e) {
    final SimpleMibDataNode node = (SimpleMibDataNode) popupPath.getLastPathComponent();
    if (node.getProxyObject() != null) {
      Thread thread = new Thread(new Runnable() {
        @Override
        public void run() {
          node.getProxyObject().reload(snmpService, target, variableBindingMapper, null);
          ((AbstractTableModel)mibTreeTable.getModel()).fireTableDataChanged();
        }
      });
      thread.start();
    }
  }

  private void doTableAction(ActionEvent e) {
    final SimpleMibDataNode node = (SimpleMibDataNode) popupPath.getLastPathComponent();

    if (node.getProxyObject() != null) {
      Thread thread = new Thread(new Runnable() {
        @Override
        public void run() {
          node.getProxyObject().reload(snmpService, target, variableBindingMapper, null);
        }
      });
      thread.start();
      final JDialog dialog = new JDialog(this);
      final SnmpTable snmpTable = (org.snmp4j.model.snmp.proxy.impl.SnmpTableImpl)node.getProxyObject();
      Graphics g = this.getGraphics();
      final int maxAdvance = g.getFontMetrics().getMaxAdvance();
      SimpleMibTableBrowser tableBrowser =
          new SimpleMibTableBrowser(snmpTable, snmpTable.getTableColumnModel(maxAdvance, 20),
              snmpService, smiManager, target, variableBindingMapper) {
        @Override
        public void close() {
          dialog.setVisible(false);
        }
      };
      dialog.setTitle(node.getObjectName());
      dialog.setLocationRelativeTo(this);
      dialog.setLayout(new BorderLayout(10,10));
      dialog.add(BorderLayout.CENTER, tableBrowser.$$$getRootComponent$$$());
      dialog.pack();
      dialog.setVisible(true);
    }
  }

  private SimpleMibDataNode createDataMibStructure() {
    SmiObject rootNode = smiManager.findSmiObject(new OID("1.3.6.1"));
    if (rootNode == null) {
      rootNode = smiManager.findRootSmiObject();
    }
    if (rootNode == null) {
      return null;
    }
    SimpleMibDataNode node = new SimpleMibDataNode(rootNode, null);
    buildSubtree(node);
    return node;
  }

  private void buildSubtree(SimpleMibDataNode node) {
    List<? extends SmiObject> children = node.getSmiObject().getChildren();
    ArrayList<SimpleMibDataNode> childNodes = new ArrayList<SimpleMibDataNode>(children.size());
    for (SmiObject child : children) {
      SnmpProxyObject proxyObject = null;
      boolean buildSubtree = true;
      if (child instanceof SmiObjectType) {
        SmiObjectType smiObjectType = (SmiObjectType) child;
        switch (child.getType()) {
          case OBJECT_TYPE_SCALAR: {
            proxyObject = new SnmpScalarImplRO<String>(child.getOID(), null, null, String.class, child.getSmiSyntax());
            buildSubtree = false;
            break;
          }
          case OBJECT_TYPE_TABLE:
            try {
              SmiObjectType tableEntry = (SmiObjectType) smiObjectType.getChildren().get(0);
              proxyObject = new SnmpTableImpl<SnmpProxyIndexColumn<String>, SnmpProxyColumn<String>, String, String,
                  SnmpProxyRowImpl<String,String>>(tableEntry.getOID(), null,
                  snmpProxyAdapter, this,
                  TableHelper.buildIndexColumns(tableEntry), TableHelper.buildColumns(tableEntry), 10);
            }
            catch (NullPointerException npe) {
              // TODO: log
            }
            catch (ClassCastException cce) {
              // TODO: log
            }
            break;
          case OBJECT_TYPE_TABLE_ENTRY: {
            proxyObject = new SnmpTableImpl<SnmpProxyIndexColumn<String>,
                SnmpProxyColumn<String>, String, String, SnmpProxyRowImpl<String, String>>(smiObjectType.getOID(), null,
                snmpProxyAdapter, this, TableHelper.buildIndexColumns(smiObjectType),
                TableHelper.buildColumns(smiObjectType), 10);
            break;
          }
        }
      }
      SimpleMibDataNode childNode = new SimpleMibDataNode(child, proxyObject);
      childNodes.add(childNode);
      if (buildSubtree) {
        buildSubtree(childNode);
      }
    }
    node.setChildren(childNodes);
  }

  private static SmiManager configureSnmp4jSMI(Map<String,List<?>> commandLineParameters) throws IOException {
    SmiManager smiManager;
    // Define the MIB repository
    String licenseKey = "";
    if (commandLineParameters.containsKey("L")) {
      licenseKey = (String) ArgumentParser.getFirstValue(commandLineParameters.get("L"));
    }
    if (commandLineParameters.containsKey("M")) {
      @SuppressWarnings(value = "unchecked")
      String f = (String) ArgumentParser.getFirstValue(commandLineParameters.get("M"));
      smiManager = new SmiManager(licenseKey, new File(f));
    }
    else {
      smiManager = new SmiManager(licenseKey, new MemRepositoryDriver());
    }
    // Load MIB modules from the repository
    if (commandLineParameters.containsKey("m")) {
      List<String> modulePatterns = new ArrayList<String>();
      for (Object p : commandLineParameters.get("m")) {
        modulePatterns.add(p.toString());
      }
      loadModules(smiManager, modulePatterns);
    }
    // Here the formatting with MIB information is finally configured in SNMP4J:
    SNMP4JSettings.setOIDTextFormat(smiManager);
    SNMP4JSettings.setVariableTextFormat(smiManager);
    smiManager.setOidFormat(SmiManager.OIDFormat.ObjectNameAndDecodedIndex4RoundTrip);
    //smiManager.setOidVariableLengthStringQuote(smiManager.getOidFixedLengthStringQuote());
    smiManager.setOctetStringDisplayHint(SmiManager.OctetStringDefaultFormat.MIB);
    return smiManager;
  }

  private static void loadModules(SmiManager smiManager, List<String> modulePatterns) throws IOException {
    String[] availableModules = smiManager.listModules();
    for (String modulePattern : modulePatterns) {
      Pattern p = Pattern.compile(modulePattern);
      boolean matched = false;
      for (String availableModule : availableModules) {
        Matcher m = p.matcher(availableModule);
        if (m.matches()) {
          matched = smiManager.loadModule(availableModule);
          System.out.println("Loaded MIB module '"+availableModule+"'");
        }
      }
      if (!matched) {
        System.err.println("No module loaded for pattern '" + modulePattern + "'");
      }
    }
  }

  public static void main(final String[] args) {
    Runnable gui = new Runnable() {

      public void run() {
        try {
          UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        } catch (Exception e) {
          e.printStackTrace();
        }
        try {
          Map<String,List<?>> commandLineParameters;
          try {
            ArgumentParser parser =
                new ArgumentParser(ALL_OPTIONS, ADDRESS_PARAMETER);
            commandLineParameters = parser.parse(args);
          new SimpleMibBrowser(commandLineParameters).setVisible(true);
          } catch (ParseException e) {
            e.printStackTrace();
          }
        } catch (IOException e) {
          e.printStackTrace();
        }
      }
    };
    SwingUtilities.invokeLater(gui);
  }

  @Override
  public SnmpProxyRowImpl<String,String> createProxyRow(OID index, List<String> indexObjects, List<String> values) {
    return new SnmpProxyRowImpl<String, String>(index, indexObjects, values);
  }
}
