/*_############################################################################
  _## 
  _##  SNMP4J-Model - VariableBindingMapper.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/

package org.snmp4j.model.mapper;

import org.snmp4j.smi.OID;
import org.snmp4j.smi.VariableBinding;

import java.text.ParseException;

/**
 * This <code>VariableBindingMapper</code> interface provides conversion methods
 * to convert {@link VariableBinding} instances to and from signed long,
 * unsigned long, string, date and time, and binary values.
 *
 * @param <L>
 *   the model type for signed long values.
 * @param <U>
 *   the model type for unsigned long values.
 * @param <S>
 *   the model type for string values.
 * @param <D>
 *   the model type for date (DateAndTime) values.
 * @param <B>
 *   the model type for binary values (OCTET STRING, BITS).
 *
 * @author Frank Fock
 * @version 1.0
 */
public interface VariableBindingMapper<L,U,S,D,B> {

  /**
   * Convert the provided variable binding to a long value.
   * @param variableBinding
   *   a SNMP4J variable binding.
   * @return
   *   a long value.
   */
  L toLong(VariableBinding variableBinding);

  /**
   * Convert the provided variable binding to an unsigned long value.
   * @param variableBinding
   *   a SNMP4J variable binding.
   * @return
   *   an unsigned long.
   */
  U toUnsignedLong(VariableBinding variableBinding);

  /**
   * Convert the provided variable binding to a string value.
   * @param variableBinding
   *   a SNMP4J variable binding.
   * @return
   *   a string value.
   */
  S toString(VariableBinding variableBinding);

  /**
   * Convert the provided variable binding to a date value.
   * @param variableBinding
   *   a SNMP4J variable binding.
   * @return
   *   a date value.
   */
  D toDate(VariableBinding variableBinding);

  /**
   * Convert the provided variable binding to a binary value.
   * @param variableBinding
   *   a SNMP4J variable binding.
   * @return
   *   a binary value.
   */
  B toBinary(VariableBinding variableBinding);

  /**
   * Convert the provided variable binding to a value of the specified class.
   * @param variableBinding
   *    the SNMP4J variable binding to convert (no modification of the object itself).
   * @param targetClass
   *    the class of the return type.
   * @return
   *    an instance of <code>targetClass</code> with the converted value of
   *    <code>variableBinding</code> or <code>null</code> if the conversion failed.
   */
  Object toObject(VariableBinding variableBinding, Class targetClass);

  /**
   * Create a SNMP4J variable binding for an {@link OID} and a long value.
   * @param oid
   *    an OID.
   * @param longValue
   *    a numeric long value.
   * @return
   *    a {@link org.snmp4j.smi.VariableBinding} instance with the {@link OID} <code>oid</code> and variable
   *    that matches the SNMP type defined for that OID with the specified long value,
   *    <code>null</code> will be returned if the conversion failed.
   */
  VariableBinding fromLong(OID oid, L longValue);

  /**
   * Create a SNMP4J variable binding for an {@link OID} and an unsigned long value.
   * @param oid
   *    an OID.
   * @param unsignedLongValue
   *    a numeric unsigned long value.
   * @return
   *    a {@link org.snmp4j.smi.VariableBinding} instance with the {@link OID} <code>oid</code> and variable
   *    that matches the SNMP type defined for that OID with the specified unsigned long value,
   *    <code>null</code> will be returned if the conversion failed.
   */
  VariableBinding fromUnsignedLong(OID oid, U unsignedLongValue);

  /**
   * Create a SNMP4J variable binding for an {@link OID} and a string value.
   * @param oid
   *    an OID.
   * @param stringValue
   *    a numeric long value.
   * @return
   *    a {@link org.snmp4j.smi.VariableBinding} instance with the {@link OID} <code>oid</code> and variable
   *    that matches the SNMP type defined for that OID with the specified string value,
   *    <code>null</code> will be returned if the conversion failed.
   */
  VariableBinding fromString(OID oid, S stringValue);

  /**
   * Create a SNMP4J variable binding for an {@link OID} and a date value.
   * @param oid
   *    an OID.
   * @param dateValue
   *    a string value.
   * @return
   *    a {@link org.snmp4j.smi.VariableBinding} instance with the {@link OID} <code>oid</code> and variable
   *    that matches the SNMP type defined for that OID with the specified date value,
   *    <code>null</code> will be returned if the conversion failed.
   */
  VariableBinding fromDate(OID oid, D dateValue);

  /**
   * Create a SNMP4J variable binding for an {@link OID} and a binary string value.
   * @param oid
   *    an OID.
   * @param binaryValue
   *    a binary string value.
   * @return
   *    a {@link org.snmp4j.smi.VariableBinding} instance with the {@link OID} <code>oid</code> and variable
   *    that matches the SNMP type defined for that OID with the specified binary value,
   *    <code>null</code> will be returned if the conversion failed.
   */
  VariableBinding fromBinary(OID oid, B binaryValue);

  /**
   * Create a SNMP4J variable binding for an {@link OID} and an object value.
   * @param oid
   *    an OID.
   * @param object
   *    an object value.
   * @return
   *    a {@link org.snmp4j.smi.VariableBinding} instance with the {@link OID} <code>oid</code> and variable
   *    that matches the SNMP type defined for that OID with the specified object value,
   *    <code>null</code> will be returned if the conversion failed.
   */
  VariableBinding fromObject(OID oid, Object object);

}
