/*_############################################################################
  _## 
  _##  SNMP4J-Model - CommitStatus.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.api;

import org.snmp4j.model.snmp.proxy.impl.SnmpValuesChangeSet;
import org.snmp4j.model.snmp.spi.SnmpCommitResult;
import org.snmp4j.model.snmp.spi.SnmpErrorStatus;


import java.util.List;

/**
 * The <code>CommitStatus</code> represents the status of a SNMP commit operation.
 * It has two lists, the first contains all the successfully committed changes and the second
 * holds any failed change sets.
 * @author Frank Fock
 * @version 1.0
 */
public class CommitStatus implements SnmpCommitResult {

  private SnmpErrorStatus snmpErrorStatus;
  private int errorIndex;
  private List<SnmpValuesChangeSet> failedChanges;
  private List<SnmpValuesChangeSet> committedChanges;

  /**
   * Create a SNMP commit result based on an error status and index as well as a list of succeeded
   * (committed) changes and failed changes.
   * @param snmpErrorStatus
   *   the SNMP error status that caused the first failure as defined by {@link SnmpErrorStatus}.
   * @param errorIndex
   *   the index of the value that failed. Counting starts from the first change set's first element (=1) until the last
   *   changes set's last value.
   * @param committedChanges
   *   the (possibly empty) list of successfully committed changes.
   * @param failedChanges
   *   the (possibly empty) list of uncommitted changes or changes that caused an error.
   */
  public CommitStatus(SnmpErrorStatus snmpErrorStatus, int errorIndex,
                      List<SnmpValuesChangeSet> committedChanges,
                      List<SnmpValuesChangeSet> failedChanges) {
    this.snmpErrorStatus = snmpErrorStatus;
    this.errorIndex = errorIndex;
    this.failedChanges = failedChanges;
    this.committedChanges = committedChanges;
  }

  @Override
  public boolean isSuccess() {
    return snmpErrorStatus == SnmpErrorStatus.noError;
  }

  @Override
  public SnmpErrorStatus getErrorStatus() {
    return snmpErrorStatus;
  }

  @Override
  public List<SnmpValuesChangeSet> getFailedChanges() {
    return failedChanges;
  }

  @Override
  public List<SnmpValuesChangeSet> getCommittedChanges() {
    return committedChanges;
  }

  @Override
  public int getErrorIndex() {
    return errorIndex;
  }

  @Override
  public String toString() {
    return "CommitStatus{" +
        "snmpErrorStatus=" + snmpErrorStatus +
        ", errorIndex=" + errorIndex +
        ", failedChanges=" + failedChanges +
        ", committedChanges=" + committedChanges +
        '}';
  }
}
