/*_############################################################################
  _## 
  _##  SNMP4J-Model - LocalSnmpService.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.api;

import org.snmp4j.*;
import org.snmp4j.event.ResponseEvent;
import org.snmp4j.log.LogAdapter;
import org.snmp4j.log.LogFactory;
import org.snmp4j.model.snmp.proxy.SnmpCommitListener;
import org.snmp4j.model.snmp.proxy.impl.SnmpValuesChangeSet;
import org.snmp4j.model.snmp.spi.*;
import org.snmp4j.model.transaction.SnmpIsolationLevel;
import org.snmp4j.model.transaction.SnmpTransaction;
import org.snmp4j.model.transaction.SnmpTransactionStrategy;
import org.snmp4j.model.transaction.TransactionManager;
import org.snmp4j.mp.MPv3;
import org.snmp4j.mp.MessageProcessingModel;
import org.snmp4j.mp.SnmpConstants;
import org.snmp4j.security.*;
import org.snmp4j.smi.OID;
import org.snmp4j.smi.VariableBinding;
import org.snmp4j.transport.DefaultUdpTransportMapping;
import org.snmp4j.util.*;

import java.io.IOException;
import java.util.*;

/**
 * The LocalSnmpService is the default {@link org.snmp4j.model.snmp.spi.SnmpService} implementation that
 * sends directly (from the local device) SNMP request to the specified targets.
 *
 * @author Frank Fock
 * @version 1.0
 */
public class LocalSnmpService implements SnmpService, TransactionManager {

  private static final LogAdapter LOGGER = LogFactory.getLogger(LocalSnmpService.class);

  private Snmp snmp;
  private PDUFactory pduFactory;
  private WeakHashMap<GlobalTransaction, Map<Target, LocalTransaction>> transactions =
      new WeakHashMap<GlobalTransaction, Map<Target, LocalTransaction>>();

  /**
   * Creates a local {@link org.snmp4j.model.snmp.spi.SnmpService} with {@link org.snmp4j.security.USM} and
   * {@link org.snmp4j.security.TSM} security models, the {@link org.snmp4j.mp.MPv3} message processing model,
   * and the {@link org.snmp4j.transport.DefaultUdpTransportMapping} transport mapping.
   */
  public LocalSnmpService() {
    this(new MessageDispatcherImpl(), null, null,
         SecurityModels.getCollection(new SecurityModel[]{ new USM(), new TSM() }),
         SecurityProtocols.getInstance().addDefaultProtocols());
  }

  /**
   * Creates a local {@link org.snmp4j.model.snmp.spi.SnmpService} with the given message dispatcher,
   * transport mappings, message processing and security models, and security protocols.
   * @param messageDispatcher
   *    the message dispatcher to be used by this local SNMP service. Cannot be <code>null</code>.
   * @param transportMappings
   *    the transport mappings to be used. If <code>null</code> is given, a new
   *    {@link org.snmp4j.transport.DefaultUdpTransportMapping} will be created and used.
   * @param messageProcessingModels
   *    the message processing models to be used by the service. If <code>null</code> is given, a new
   *    {@link org.snmp4j.mp.MPv3} is created and used (without any SNMPv1 or v2c models).
   * @param securityModels
   *    the security models to be used with the provided message processing models. If there is no
   *    message processing model, that requires a security model (i.e., {@link org.snmp4j.mp.MPv1} and
   *    {@link org.snmp4j.mp.MPv2c} only) then this value can be <code>null</code>.
   * @param securityProtocols
   *    the security protocols to be used with the provided message processing models. If there is no
   *    message processing model, that requires a security protocol (i.e., {@link org.snmp4j.mp.MPv1} and
   *    {@link org.snmp4j.mp.MPv2c} only) then this value can be <code>null</code>.
   *
   */
  public LocalSnmpService(MessageDispatcher messageDispatcher,
                          List<TransportMapping> transportMappings,
                          List<MessageProcessingModel> messageProcessingModels,
                          SecurityModels securityModels,
                          SecurityProtocols securityProtocols) {
    snmp = new Snmp(messageDispatcher);
    if (messageProcessingModels == null) {
      messageProcessingModels = Arrays.asList(new MessageProcessingModel[] { new MPv3() });
    }
    if (transportMappings == null) {
      try {
        transportMappings = Arrays.asList(new TransportMapping[] { new DefaultUdpTransportMapping() });
      } catch (IOException e) {
        LOGGER.error("Failed to create default UDP transport mapping: "+e.getMessage(), e);
      }
    }
    for (MessageProcessingModel mp : messageProcessingModels) {
      messageDispatcher.addMessageProcessingModel(mp);
      if (mp instanceof MPv3) {
        MPv3 mpv3 = (MPv3)mp;
        mpv3.setSecurityModels(securityModels);
        mpv3.setSecurityProtocols(securityProtocols);
      }
    }
    for (TransportMapping tm : transportMappings) {
      messageDispatcher.addTransportMapping(tm);
      tm.addTransportListener(messageDispatcher);
    }
    this.pduFactory = new DefaultPDUFactory(PDU.GET);
    LOGGER.info("LocalSnmpService initialized. TransportMappings="+transportMappings+", MPs="+messageProcessingModels);
  }

  /**
   * Returns the {@link org.snmp4j.Snmp} instance used by this local {@link org.snmp4j.model.snmp.spi.SnmpService}.
   * @return
   *    a {@link org.snmp4j.Snmp} instance.
   */
  public Snmp getSnmp() {
    return snmp;
  }

  @Override
  public void initTransports() throws IOException {
    snmp.listen();
  }

  /**
   * Gets the {@link org.snmp4j.util.PDUFactory} of this SNMP service.
   * @return
   *    a {@link org.snmp4j.util.PDUFactory} instance.
   */
  public PDUFactory getPduFactory() {
    return pduFactory;
  }

  /**
   * Sets the {@link org.snmp4j.util.PDUFactory} of this SNMP service. The factory is used to create any SNMP
   * {@link org.snmp4j.PDU} created on behalf of this service. By default, this is the
   * {@link org.snmp4j.util.DefaultPDUFactory} with the PDU type {@link PDU#GET}.
   *
   * @param pduFactory
   *    the {@link org.snmp4j.util.PDUFactory} to be used by this local SNMP service.
   */
  public void setPduFactory(PDUFactory pduFactory) {
    this.pduFactory = pduFactory;
  }

  @Override
  public List<? extends VariableBinding> get(Target target, List<OID> oidList) throws SnmpException {
    PDU pdu = pduFactory.createPDU(target);
    ArrayList<VariableBinding> vbs = new ArrayList<VariableBinding>(oidList.size());
    for (OID oid : oidList) {
      vbs.add(new VariableBinding(oid));
    }
    pdu.setVariableBindings(vbs);
    pdu.setRequestID(null);
    try {
      ResponseEvent responseEvent = snmp.get(pdu, target);
      if (responseEvent.getResponse() != null) {
        if (responseEvent.getResponse().getErrorStatus() == SnmpConstants.SNMP_ERROR_SUCCESS) {
          return responseEvent.getResponse().getVariableBindings();
        }
        else {
          throw new SnmpException(SnmpErrorStatus.fromSnmpErrorStatus(responseEvent.getResponse().getErrorStatus()));
        }
      }
      else {
        throw new SnmpException("Request timed out", responseEvent.getError(), SnmpErrorStatus.timeout);
      }
    } catch (IOException e) {
      e.printStackTrace();
      throw new SnmpException("IO exception", e, null);
    }
  }

  @Override
  public List<? extends VariableBinding> getNext(Target target, List<OID> oidList) throws SnmpException {
    PDU pdu = pduFactory.createPDU(target);
    ArrayList<VariableBinding> vbs = new ArrayList<VariableBinding>(oidList.size());
    for (OID oid : oidList) {
      vbs.add(new VariableBinding(oid));
    }
    pdu.setVariableBindings(vbs);
    try {
      ResponseEvent responseEvent = snmp.getNext(pdu, target);
      if (responseEvent.getResponse() != null) {
        if (responseEvent.getResponse().getErrorStatus() == SnmpConstants.SNMP_ERROR_SUCCESS) {
          return responseEvent.getResponse().getVariableBindings();
        }
        else {
          throw new SnmpException(SnmpErrorStatus.fromSnmpErrorStatus(responseEvent.getResponse().getErrorStatus()));
        }
      }
      else {
        throw new SnmpException("Request timed out", responseEvent.getError(), SnmpErrorStatus.timeout);
      }
    } catch (IOException e) {
      e.printStackTrace();
      throw new SnmpException("IO exception", e, null);
    }
  }

  @Override
  public List<? extends VariableBinding> getSubTree(Target target, OID oid) throws SnmpException {
    TreeUtils treeUtils = new TreeUtils(snmp, pduFactory);
    List<TreeEvent> treeEvents = treeUtils.getSubtree(target, oid);
    List<VariableBinding> instances = new ArrayList<VariableBinding>(treeEvents.size());
    for (TreeEvent treeEvent : treeEvents) {
      if (treeEvent.getStatus() == PDU.noError) {
        instances.addAll(Arrays.asList(treeEvent.getVariableBindings()));
      }
      else {
        throw new SnmpException(SnmpErrorStatus.fromSnmpErrorStatus(treeEvent.getStatus()));
      }
    }
    return instances;
  }

  @Override
  public List<? extends SnmpRow> getTable(Target target, List<OID> columnOIDs, OID minIndexOID, OID maxIndexOID)
      throws SnmpException
  {
    TableUtils tableUtils = new TableUtils(snmp, pduFactory);
    List<TableEvent> tableEvents =
        tableUtils.getTable(target, columnOIDs.toArray(new OID[columnOIDs.size()]), minIndexOID, maxIndexOID);
    List<SnmpRow> rows = new ArrayList<SnmpRow>(tableEvents.size());
    for (TableEvent tableEvent : tableEvents) {
      if (tableEvent.getStatus() == PDU.noError) {
        SnmpRow row = new SnmpRow(tableEvent.getIndex(), Arrays.asList(tableEvent.getColumns()));
        rows.add(row);
      }
      else {
        throw new SnmpException(SnmpErrorStatus.fromSnmpErrorStatus(tableEvent.getStatus()));
      }
    }
    return rows;
  }

  @Override
  public void getTable(Target target, List<OID> columnOIDs, OID minIndex, OID maxIndex,
                       SnmpRowListener callback, Object userObject) throws SnmpException {
    TableUtils tableUtils = new TableUtils(snmp, pduFactory);
    TableListener tableListener = new LocalTableListener(callback);
    tableUtils.getTable(target, columnOIDs.toArray(new OID[columnOIDs.size()]),
                        tableListener, userObject, minIndex, maxIndex);
  }

  @Override
  public void update(SnmpTransaction snmpTransaction, Target target,
                     SnmpValuesChangeSet variableBindings, SnmpCommitListener commitListener)  {
    Map<Target,LocalTransaction> localTransactions = transactions.get(snmpTransaction);
    if (localTransactions == null) {
      throw new IllegalStateException("Transaction not available: "+snmpTransaction);
    }
    LocalTransaction localTransaction = localTransactions.get(target);
    if (localTransaction == null) {
      localTransaction = new LocalTransaction(target);
      localTransactions.put(target, localTransaction);
    }
    localTransaction.addUpdate(variableBindings);
    if (snmpTransaction.isAutoCommit()) {
      commit(snmpTransaction, commitListener);
    }
  }

  @Override
  public List<? extends SnmpCommitResult> commit(SnmpTransaction transaction, SnmpCommitListener commitListener) {
    Map<Target,LocalTransaction> localTransactions = transactions.get(transaction);
    if (localTransactions == null) {
      throw new IllegalStateException("Transaction not available: "+transaction);
    }
    ArrayList<CommitStatus> commitResults = new ArrayList<CommitStatus>();
    for (Map.Entry<Target, LocalTransaction> localTransactionEntry : localTransactions.entrySet()) {
      LocalTransaction localTransaction = localTransactionEntry.getValue();
      Queue<SnmpValuesChangeSet> updates = localTransaction.getPending();
      List<SnmpValuesChangeSet> localTransactionPending =
          new ArrayList<SnmpValuesChangeSet>();
      int maxSizePDU = localTransaction.getTarget().getMaxSizeRequestPDU();
      PDU pdu = pduFactory.createPDU(localTransaction.getTarget());
      if (pdu instanceof ScopedPDU) {
        ((ScopedPDU)pdu).setContextName(commitListener.getContext());
      }
      SnmpErrorStatus snmpErrorStatus = SnmpErrorStatus.noError;
      while ((snmpErrorStatus == SnmpErrorStatus.noError) &&
          (!updates.isEmpty()) && localTransaction.getFailedVariableBindings().isEmpty()) {
        SnmpValuesChangeSet nextUpdate = updates.poll();
        pdu.addAll(nextUpdate.getVariableBindings());
        localTransactionPending.add(nextUpdate);
        if (!updates.isEmpty()) {
          SnmpValuesChangeSet successorUpdate = updates.peek();
          if ((successorUpdate != null) &&
              ((pdu.getBERLength() + PDU.getBERLength(successorUpdate.getVariableBindings())) < maxSizePDU)) {
            continue;
          }
        }
        // PDU is complete
        try {
          ResponseEvent responseEvent = snmp.set(pdu, localTransaction.getTarget());
          PDU responsePDU = responseEvent.getResponse();
          CommitStatus commitStatus;
          if (responsePDU != null) {
            if (responsePDU.getErrorStatus() == PDU.noError) {
              commitStatus = new CommitStatus(SnmpErrorStatus.noError, 0, localTransactionPending, null);
              localTransaction.getCommitted().addAll(localTransactionPending);
              if (commitListener != null) {
                commitListener.commitSuccess(localTransactionPending);
              }
            }
            else {
              snmpErrorStatus = SnmpErrorStatus.fromSnmpErrorStatus(responsePDU.getErrorStatus());
              localTransaction.setErrorIndex(responsePDU.getErrorIndex());
              localTransaction.getFailedVariableBindings().addAll(pdu.getVariableBindings());
              commitStatus = new CommitStatus(snmpErrorStatus, responsePDU.getErrorIndex(), null, localTransactionPending);
              if (commitListener != null) {
                commitListener.commitFailure(localTransactionPending, snmpErrorStatus, localTransaction.getErrorIndex());
              }
            }
          }
          else {
            localTransaction.getFailedVariableBindings().addAll(pdu.getVariableBindings());
            snmpErrorStatus = SnmpErrorStatus.timeout;
            commitStatus = new CommitStatus(snmpErrorStatus, 0, null, localTransactionPending);
            if (commitListener != null) {
              commitListener.commitFailure(localTransactionPending, snmpErrorStatus, 0);
            }
          }
          commitResults.add(commitStatus);
        } catch (IOException e) {
          LOGGER.error("IOException occurred while committing transaction '" + localTransaction + "': " + e.getMessage(), e);
          localTransaction.getFailedVariableBindings().addAll(pdu.getVariableBindings());
          snmpErrorStatus = SnmpErrorStatus.ioError;
        }
        // Create next PDU
        pdu = pduFactory.createPDU(localTransaction.getTarget());
      }
      localTransaction.setSnmpErrorStatus(snmpErrorStatus);
    }
    return commitResults;
  }

  @Override
  public void rollback(SnmpTransaction transaction) {
    Map<Target,LocalTransaction> localTransactions = transactions.get(transaction);
    if (localTransactions != null) {
      for (LocalTransaction localTransaction : localTransactions.values()) {
        localTransaction.getPending().clear();
      }
    }
  }

  @Override
  public boolean isAtomic(SnmpTransaction transaction) {
    Map<Target,LocalTransaction> localTransactions = transactions.get(transaction);
    if (localTransactions != null) {
//      if (localTransactions.)
    }
    return true;
  }

  @Override
  public SnmpTransaction getSnmpTransaction(SnmpTransactionStrategy transactionStrategy) {
    GlobalTransaction globalTransaction =
        new GlobalTransaction(SnmpIsolationLevel.none == transactionStrategy.getIsolationLevel());
    transactions.put(globalTransaction, new LinkedHashMap<Target, LocalTransaction>());
    return globalTransaction;
  }

  protected class LocalTableListener implements TableListener {

    private SnmpRowListener rowListener;
    private boolean finished;
    private boolean initialized;

    public LocalTableListener(SnmpRowListener rowListener) {
      this.rowListener = rowListener;
    }

    @Override
    public boolean next(TableEvent tableEvent) {
      if (!initialized) {
        rowListener.beginUpdate(tableEvent.getUserObject());
        initialized = true;
      }
      if (tableEvent.getStatus() == PDU.noError) {
        SnmpRow row = new SnmpRow(tableEvent.getIndex(), Arrays.asList(tableEvent.getColumns()));
        return rowListener.nextRow(row, tableEvent.getUserObject());
      }
      else {
        rowListener.endUpdate(SnmpErrorStatus.fromSnmpErrorStatus(tableEvent.getStatus()), tableEvent.getUserObject());
        finished = true;
      }
      return false;
    }

    @Override
    public void finished(TableEvent event) {
      rowListener.endUpdate(SnmpErrorStatus.fromSnmpErrorStatus(event.getStatus()), event.getUserObject());
      finished = true;
    }

    @Override
    public boolean isFinished() {
      return finished;
    }


  }


}
