/*_############################################################################
  _## 
  _##  SNMP4J-Model - LocalTransaction.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.api;

import org.snmp4j.Target;
import org.snmp4j.model.snmp.proxy.impl.SnmpValuesChangeSet;
import org.snmp4j.model.snmp.spi.SnmpErrorStatus;
import org.snmp4j.smi.VariableBinding;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedDeque;

/**
 * The <code>LocalTransaction</code> holds changes, error status, and target information
 * for a SNMP transaction.
 * @author Frank Fock
 * @version 1.0
 */
public class LocalTransaction {

  private Target target;
  private Queue<SnmpValuesChangeSet> pending =
      new ConcurrentLinkedDeque<SnmpValuesChangeSet>();
  private List<VariableBinding> failed = new ArrayList<VariableBinding>(1);
  private SnmpErrorStatus snmpErrorStatus;
  private int errorIndex = 0;
  private List<SnmpValuesChangeSet> committed =
      Collections.synchronizedList(new ArrayList<SnmpValuesChangeSet>());

  /**
   * Create a new transaction for a specific SNMP {@link org.snmp4j.Target}.
   * @param target
   *    the target SNMP entity subject for this transaction.
   */
  public LocalTransaction(Target target) {
    this.target = target;
  }

  /**
   * Gets the SNMP target of this transaction.
   * @return
   *    a {@link org.snmp4j.Target} instance.
   */
  public Target getTarget() {
    return target;
  }

  /**
   * Gets the queue of pending {@link org.snmp4j.model.snmp.proxy.impl.SnmpValuesChangeSet}s.
   * @return
   *    the queue of pending value changes. If empty, all changes have been committed (or there no changes in the
   *    transaction yet).
   */
  public Queue<SnmpValuesChangeSet> getPending() {
    return pending;
  }

  /**
   * Gets the list of (successfully) committed changes.
   * @return
   *    a list of {@link org.snmp4j.model.snmp.proxy.impl.SnmpValuesChangeSet} instances. If empty, no changes
   *    have been successfully committed yet. Failed changes will be reported by {@link #getFailedVariableBindings()}.
   */
  public List<SnmpValuesChangeSet> getCommitted() {
    return committed;
  }

  /**
   * Adds a value change set to this transaction. The change set is appended to the queue of pending change sets.
   * @param changeSet
   *    a value change set to include in a future commit.
   */
  public void addUpdate(SnmpValuesChangeSet changeSet) {
    this.pending.add(changeSet);
  }

  /**
   * After this transaction completed, this method indicates whether there a failed change sets or not.
   * @return
   *    <code>true</code> if there are failed value change sets.
   */
  public boolean isSuccess() {
    return failed.isEmpty();
  }

  /**
   * Gets the SNMP error status of the transaction.
   * @return
   *    a {@link org.snmp4j.model.snmp.spi.SnmpErrorStatus} instance that represents the status of the SNMP operation(s)
   *    used to execute this transaction.
   */
  public SnmpErrorStatus getErrorStatus() {
    return snmpErrorStatus;
  }

  /**
   * Gets the list of failed variable bindings (new values that could not applied to an error).
   * @return
   *    a possibly empty list of {@link org.snmp4j.smi.VariableBinding} instances.
   */
  public List<VariableBinding> getFailedVariableBindings() {
    return failed;
  }

  /**
   *
   * @return
   */
  public int getErrorIndex() {
    return errorIndex;
  }

  public void setErrorIndex(int errorIndex) {
    this.errorIndex = errorIndex;
  }

  public void setSnmpErrorStatus(SnmpErrorStatus snmpErrorStatus) {
    this.snmpErrorStatus = snmpErrorStatus;
  }

  @Override
  public String toString() {
    return "LocalTransaction{" +
        "target=" + target +
        ", pending=" + pending +
        ", failed=" + failed +
        ", snmpErrorStatus=" + snmpErrorStatus +
        ", errorIndex=" + errorIndex +
        ", committed=" + committed +
        '}';
  }
}
