/*_############################################################################
  _## 
  _##  SNMP4J-Model - SnmpColumn.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.proxy.impl;

import org.snmp4j.model.snmp.proxy.SnmpProxyColumn;
import org.snmp4j.smi.MaxAccess;
import org.snmp4j.smi.OID;

/**
 * The SnmpColumn class represents a columnar MIB object.
 *
 * @param <T>
 *   specifies the value type of the column.
 *
 * @author Frank Fock
 */
public class SnmpColumn<T> implements SnmpProxyColumn<T> {

  private OID oid;
  private Class<T> valueClass;
  private int minWidth = -1;
  private int maxWidth = -1;
  private String header;
  private MaxAccess smiMaxAccess;
  private int smiSyntax;

  /**
   * Create a column with an OID (i.e., the object identifier of the columnar MIB object without any
   * index), value class, maximum supported access and SMI syntax.
   * @param oid
   *    the object identifier of the columnar MIB object. For example, for the column ifDescr this would be
   *    <code>new OID(1.3.6.1.2.1.2.2.1.2)</code>.
   * @param valueClass
   *    the class of the value representation for this column within the model.
   * @param smiMaxAccess
   *    the maximum access that should be supported by the model for this column (the access right should be
   *    less or equal to the maximum access right defined by the corresponding MIB).
   * @param smiSyntax
   *    the SMI syntax of the object.
   */
  public SnmpColumn(OID oid, Class<T> valueClass, MaxAccess smiMaxAccess, int smiSyntax) {
    this.valueClass = valueClass;
    this.oid = oid;
    this.smiMaxAccess = smiMaxAccess;
    this.smiSyntax = smiSyntax;
  }

  /**
   * Create a column with an OID (i.e., the object identifier of the columnar MIB object without any
   * index), value class, maximum supported access and SMI syntax.
   * @param oid
   *    the object identifier of the columnar MIB object. For example, for the column ifDescr this would be
   *    <code>new OID(1.3.6.1.2.1.2.2.1.2)</code>.
   * @param valueClass
   *    the class of the value representation for this column within the model.
   * @param smiMaxAccess
   *    the maximum access that should be supported by the model for this column (the access right should be
   *    less or equal to the maximum access right defined by the corresponding MIB).
   * @param smiSyntax
   *    the SMI syntax of the object.
   * @param minWidth
   *    the minimum width in characters of the values for this column. This servers as a hint
   *    for the column size calculation when this model is being rendered for display.
   * @param maxWidth
   *    the maximum width in characters of the values of this column. This servers as a hint
   *    for the column size calculation when this model is being rendered for display.
   */
  public SnmpColumn(OID oid, Class<T> valueClass, MaxAccess smiMaxAccess, int smiSyntax,
                    int minWidth, int maxWidth) {
    this(oid, valueClass, smiMaxAccess, smiSyntax);
    this.minWidth = minWidth;
    this.maxWidth = maxWidth;
  }

  /**
   * Create a column with an OID (i.e., the object identifier of the columnar MIB object without any
   * index), value class, maximum supported access and SMI syntax.
   * @param oid
   *    the object identifier of the columnar MIB object. For example, for the column ifDescr this would be
   *    <code>new OID(1.3.6.1.2.1.2.2.1.2)</code>.
   * @param valueClass
   *    the class of the value representation for this column within the model.
   * @param smiMaxAccess
   *    the maximum access that should be supported by the model for this column (the access right should be
   *    less or equal to the maximum access right defined by the corresponding MIB).
   * @param smiSyntax
   *    the SMI syntax of the object.
   * @param minWidth
   *    the minimum width in characters of the values for this column. This servers as a hint
   *    for the column size calculation when this model is being rendered for display.
   * @param maxWidth
   *    the maximum width in characters of the values of this column. This servers as a hint
   *    for the column size calculation when this model is being rendered for display.
   * @param header
   *    the header string for this column (this value is not actually used by the proxy but can be used
   *    by a column renderer.
   */
  public SnmpColumn(OID oid, Class<T> valueClass, MaxAccess smiMaxAccess, int smiSyntax,
                    int minWidth, int maxWidth, String header) {
    this(oid, valueClass, smiMaxAccess, smiSyntax, minWidth, maxWidth);
    this.header = header;
  }

  @Override
  public OID getOID() {
    return oid;
  }

  @Override
  public Class<T> getValueClass() {
    return valueClass;
  }

  /**
   * Gets the minimum width of the column's values in characters.
   * @return
   *    the minimum character width.
   */
  public int getMinWidth() {
    return minWidth;
  }

  /**
   * Gets the maximum width of the column's values in characters.
   * @return
   *    the maximum character width.
   */
  public int getMaxWidth() {
    return maxWidth;
  }

  /**
   * Sets the column header.
   * @param header
   *    the header string for this column.
   */
  public void setHeader(String header) {
    this.header = header;
  }

  @Override
  public String getHeader() {
    return this.header;
  }

  @Override
  public MaxAccess getSmiMaxAccess() {
    return smiMaxAccess;
  }

  /**
   * Gets the SMI syntax for this column.
   * @return
   *    the SmiSyntax that represents the base MIB syntax of this column.
   */
  public int getSmiSyntax() {
    return smiSyntax;
  }

  @Override
  public String toString() {
    return "SnmpColumn{" +
        "oid=" + oid +
        ", valueClass=" + valueClass +
        ", minWidth=" + minWidth +
        ", maxWidth=" + maxWidth +
        ", header='" + header + '\'' +
        ", smiMaxAccess=" + smiMaxAccess +
        ", smiSyntax=" + smiSyntax +
        '}';
  }
}
