/*_############################################################################
  _## 
  _##  SNMP4J-Model - SnmpProxyAdapter.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.proxy.impl;

import org.snmp4j.Target;
import org.snmp4j.log.LogAdapter;
import org.snmp4j.log.LogFactory;
import org.snmp4j.model.mapper.VariableBindingMapper;
import org.snmp4j.model.snmp.proxy.*;
import org.snmp4j.model.snmp.spi.*;
import org.snmp4j.model.transaction.SnmpIsolationLevel;
import org.snmp4j.model.transaction.SnmpTransaction;
import org.snmp4j.model.transaction.SnmpTransactionStrategy;
import org.snmp4j.smi.OctetString;
import org.snmp4j.smi.VariableBinding;

import java.util.*;
import java.util.concurrent.ConcurrentLinkedQueue;

/**
 * The <code>SnmpProxyAdapter</code> is the central implementation of the mapping between the
 * {@link SnmpService} and a {@link SnmpProxyObject}. The adapter is responsible to map changes
 * collection by the proxy objects to change sets queued on the SNMP service. In addition,
 * the adapter provides the mapping of the SNMP values and the corresponding model values.
 * For read operations, this mapping is actually handled by the proxy objects themselves using
 * the {}@link VariableBindingMapper} provided by this adapter.
 *
 * @author Frank Fock
 */
public class SnmpProxyAdapter implements SnmpProxyCommitter, SnmpProxyReloader, SnmpTransactionStrategy {

  private static final LogAdapter LOG = LogFactory.getLogger(SnmpProxyAdapter.class);

  private SnmpService snmpService;
  private Target target;

  private VariableBindingMapper variableBindingMapper;

  private Map<SnmpProxyObject, List<SnmpValuesChangeSet>> commitQueue =
      new LinkedHashMap<SnmpProxyObject, List<SnmpValuesChangeSet>>();

  private Queue<ReloadTask> reloadQueue = new ConcurrentLinkedQueue<ReloadTask>();

  /**
   * Creates a proxy object adapter using a {@link SnmpService}, a SNMP {@link Target}, and a
   * {@link VariableBindingMapper}.
   * @param snmpService
   *    the SNMP service to use for reloading and updating SNMP values.
   * @param target
   *    the target (i.e., address of the SNMP agent) that act as SNMP command responder.
   * @param variableBindingMapper
   *    the variable binding mapper that maps between model values and SNMP values and vice versa.
   */
  public SnmpProxyAdapter(SnmpService snmpService, Target target, VariableBindingMapper variableBindingMapper) {
    this.snmpService = snmpService;
    this.target = target;
    this.variableBindingMapper = variableBindingMapper;
  }

  @Override
  public SnmpValuesChangeSet validateChange(SnmpValuesChangeSet change) {
    SnmpValuesChangeSet result = new SnmpValuesChangeSet();
    result.setStatus(ChangeSet.ChangeStatus.PREPARED);
    for (SnmpValueChange valueChange : change.getChanges()) {
      try {
        VariableBinding vb = variableBindingMapper.fromObject(valueChange.getOid(), valueChange.getNewValue());
        if (vb == null) {
          SnmpValueChange validatedChange = new SnmpValueChange(valueChange.getOid(), valueChange.getValueClass(),
              valueChange.getOldValue(), valueChange.getOldValue());
          validatedChange.setErrorStatus(SnmpErrorStatus.badValue);
          validatedChange.setStatus(ChangeSet.ChangeStatus.FAILURE);
          result.addChange(validatedChange);
          result.setStatus(ChangeSet.ChangeStatus.FAILURE);
        }
        else {
          result.addChange(new SnmpValueChange(valueChange.getOid(), valueChange.getValueClass(),
              variableBindingMapper.toObject(vb, valueChange.getValueClass()), valueChange.getOldValue()));
        }
      }
      catch (Exception ex) {
        LOG.error("Validation of change "+change+" failed with exception: "+ex.getMessage(), ex);
        SnmpValueChange validatedChange = new SnmpValueChange(valueChange.getOid(), valueChange.getValueClass(),
            valueChange.getOldValue(), valueChange.getOldValue());
        validatedChange.setErrorStatus(SnmpErrorStatus.badValue);
        result.addChange(validatedChange);
        result.setStatus(ChangeSet.ChangeStatus.FAILURE);
      }
    }
    return result;
  }

  @Override
  public void addChange(SnmpProxyObject proxyObject,
                        SnmpValuesChangeSet change) {
    List<SnmpValuesChangeSet> proxyCommitQueue = commitQueue.get(proxyObject);
    if (proxyCommitQueue == null) {
      proxyCommitQueue = new ArrayList<SnmpValuesChangeSet>();
      proxyCommitQueue.add(change);
      commitQueue.put(proxyObject, proxyCommitQueue);
    }
    else {
      proxyCommitQueue.add(change);
    }
  }

  @Override
  public SnmpValuesChangeSet removeLastChangeSet(SnmpProxyObject proxyObject) {
    List<SnmpValuesChangeSet> proxyCommitQueue = commitQueue.get(proxyObject);
    if (proxyCommitQueue != null && !proxyCommitQueue.isEmpty()) {
      synchronized (proxyCommitQueue) {
        if (!proxyCommitQueue.isEmpty()) {
          return proxyCommitQueue.remove(proxyCommitQueue.size()-1);
        }
      }
    }
    return null;
  }

  @Override
  public List<SnmpValuesChangeSet> removeAllChanges(SnmpProxyObject proxyObject) {
    return commitQueue.remove(proxyObject);
  }

  @Override
  public TaskId commitChanges(SnmpCommitListener commitListener) {
    CommitTask commitTask = new CommitTask(commitQueue, commitListener);
    commitTask.run();
    return commitTask;
  }

  @Override
  public int getNumberOfPendingChanges() {
    int changeCount = 0;
    for (List<SnmpValuesChangeSet> proxyCommitQueue : commitQueue.values()) {
      if (proxyCommitQueue != null && !proxyCommitQueue.isEmpty()) {
        synchronized (proxyCommitQueue) {
          if (!proxyCommitQueue.isEmpty()) {
            changeCount += proxyCommitQueue.size();
          }
        }
      }
    }
    return changeCount;
  }

  @Override
  public void queueReload(SnmpProxyObject proxyObject) {
    reloadQueue.add(new ReloadTask(proxyObject));
  }

  @Override
  public List<TaskId> reloadQueued(SnmpProxyReloadListener reloadListener) {
    return processReloadQueue(reloadListener);
  }

  protected List<TaskId> processReloadQueue(SnmpProxyReloadListener reloadListener) {
    ArrayList<TaskId> taskIds = new ArrayList<TaskId>(reloadQueue.size());
    while (reloadQueue.size()>0) {
      ReloadTask reloadTask = reloadQueue.poll();
      reloadTask.reloadListener = reloadListener;
      reloadTask.run();
      taskIds.add(reloadTask.getTaskId());
    }
    return taskIds;
  }

  @Override
  public SnmpIsolationLevel getIsolationLevel() {
    return SnmpIsolationLevel.fineGrained;
  }

  protected class ReloadTask implements Runnable {

    private SnmpProxyObject snmpProxyObject;
    private SnmpProxyReloadListener reloadListener;
    private TaskId taskId;

    private ReloadTask(SnmpProxyObject snmpProxyObject) {
      this.snmpProxyObject = snmpProxyObject;
    }

    public SnmpProxyObject getSnmpProxyObject() {
      return snmpProxyObject;
    }

    public void run() {
      taskId = snmpProxyObject.reload(snmpService, target, variableBindingMapper, reloadListener);
    }

    public TaskId getTaskId() {
      return taskId;
    }
  }

  protected class CommitTask implements Runnable, TaskId {

    private Map<SnmpProxyObject, List<SnmpValuesChangeSet>> commits;
    private SnmpCommitListener commitListener;
    private volatile boolean finished;

    public CommitTask(Map<SnmpProxyObject, List<SnmpValuesChangeSet>> commits,
                      SnmpCommitListener commitListener) {
      this.commits = commits;
      this.commitListener = commitListener;
    }


    @Override
    public void run() {
      SnmpTransaction snmpTransaction = snmpService.getSnmpTransaction(SnmpProxyAdapter.this);
      LOG.debug("Committing SNMP transaction "+snmpTransaction);
      for (Map.Entry<SnmpProxyObject, List<SnmpValuesChangeSet>> commit : commits.entrySet()) {
        SnmpProxyObject proxyObject = commit.getKey();
        CommitListenerProxy commitListenerProxy =
            new CommitListenerProxy((proxyObject instanceof SnmpCommitListener) ? (SnmpCommitListener)proxyObject : null);
        List<SnmpValuesChangeSet> changeSetList = commit.getValue();
        for (SnmpValuesChangeSet changeSet : changeSetList) {
          SnmpProxyObjectHelper.applyNewValues(variableBindingMapper, changeSet);
          LOG.debug("Committing change set "+changeSet+" to target "+target);
          snmpService.update(snmpTransaction, target, changeSet, commitListenerProxy);
        }
      }
      snmpService.commit(snmpTransaction, commitListener);
      this.finished = true;
    }

    @Override
    public boolean isFinished() {
      return finished;
    }
  }

  protected class CommitListenerProxy implements SnmpCommitListener {
    private SnmpCommitListener snmpProxyObject;

    public CommitListenerProxy(SnmpCommitListener snmpProxyObject) {
      this.snmpProxyObject = snmpProxyObject;
    }

    @Override
    public void commitSuccess(List<SnmpValuesChangeSet> committedChanges) {
      if (snmpProxyObject != null) {
        snmpProxyObject.commitSuccess(committedChanges);
      }
      commitQueue.remove(snmpProxyObject);
    }

    @Override
    public void commitFailure(List<SnmpValuesChangeSet> failedChanges, SnmpErrorStatus errorStatus, int errorIndex) {
      if (snmpProxyObject != null) {
        snmpProxyObject.commitFailure(failedChanges, errorStatus, errorIndex);
      }
    }

    @Override
    public OctetString getContext() {
      if (snmpProxyObject == null) {
        return null;
      }
      return snmpProxyObject.getContext();
    }
  }
}
