/*_############################################################################
  _## 
  _##  SNMP4J-Model - SnmpProxyObjectHelper.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.proxy.impl;

import org.snmp4j.model.mapper.VariableBindingMapper;
import org.snmp4j.model.snmp.proxy.SnmpProxyColumn;
import org.snmp4j.smi.OID;
import org.snmp4j.smi.VariableBinding;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * The <code>SnmpProxyObjectHelper</code> is an utility class that helps map model values to SNMP and vice versa.
 *
 * @author Frank Fock
 */
public abstract class SnmpProxyObjectHelper {

  /**
   * Convert a list of objects from their model representation to SNMP variable bindings.
   * @param variableBindingMapper
   *    the variable binding mapper used for the conversion.
   * @param oids
   *    the object identifiers of the SNMP object types to convert (the n-th OID is associated with the n-th object
   *    of <code>objects</code>.
   * @param objects
   *    the mode object values. The type of the object must match the mapper type for that OID.
   * @return
   *    a list of variable bindings with the same size as <code>oids</code>.
   */
  public static List<VariableBinding> mapObjectsToVB(VariableBindingMapper variableBindingMapper,
                                                     List<OID> oids, List<Object> objects) {
    List<VariableBinding> vbs = new ArrayList<VariableBinding>(Math.min(oids.size(), objects.size()));
    for (Iterator oit = oids.iterator(), vit = objects.iterator();
         oit.hasNext() && vit.hasNext(); oit.next(),vit.next()) {
      vbs.add(variableBindingMapper.fromObject((OID)oit.next(), vit.next()));
    }
    return vbs;
  }

  /**
   * Convert a list of objects from their model representation to SNMP variable bindings.
   * @param variableBindingMapper
   *    the variable binding mapper used for the conversion.
   * @param vbs
   *    a list of SNMP variable bindings to convert.
   * @param targetColumns
   *    a list of columnar objects that represent the target model structure. The length of this list must match the
   *    length of <code>vbs</code> list!
   * @return
   *    a list of variable bindings with the same size as <code>oids</code>.
   */
  public static <Type> List<Type> mapVBsToObjects(VariableBindingMapper variableBindingMapper,
                                             List<VariableBinding> vbs, List<? extends SnmpProxyColumn> targetColumns) {
    ArrayList<Type> objects = new ArrayList<Type>(vbs.size());
    for (int i=0; i<vbs.size() && i<targetColumns.size(); i++) {
      objects.add((Type)variableBindingMapper.toObject(vbs.get(i), targetColumns.get(i).getValueClass()));
    }
    return objects;
  }

  /**
   * Convert any new model values in a change set to the corresponding SNMP value and store that SNMP value in the
   * change set for further processing.
   *
   * @param variableBindingMapper
   *    the variable binding mapper used for the conversion.
   * @param changeSet
   *    the change set to be modified.
   */
  public static void applyNewValues(VariableBindingMapper variableBindingMapper, SnmpValuesChangeSet changeSet) {
    for (SnmpValueChange change : changeSet.getChanges()) {
      VariableBinding vb = variableBindingMapper.fromObject(change.getOid(), change.getNewValue());
      change.setVariable(vb.getVariable());
    }
  }

  /**
   * Convert any old model values in a change set to the corresponding SNMP value and store that SNMP value in the
   * change set for further processing.
   *
   * @param variableBindingMapper
   *    the variable binding mapper used for the conversion.
   * @param changeSet
   *    the change set to be modified.
   */
  public static void applyOldValues(VariableBindingMapper variableBindingMapper, SnmpValuesChangeSet changeSet) {
    for (SnmpValueChange change : changeSet.getChanges()) {
      VariableBinding vb = variableBindingMapper.fromObject(change.getOid(), change.getOldValue());
      change.setVariable(vb.getVariable());
    }
  }

}
