/*_############################################################################
  _## 
  _##  SNMP4J-Model - SnmpScalarImpl.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.proxy.impl;

import org.snmp4j.model.snmp.proxy.SnmpCommitListener;
import org.snmp4j.model.snmp.proxy.SnmpScalar;
import org.snmp4j.model.snmp.spi.SnmpProxyCommitter;
import org.snmp4j.model.snmp.spi.SnmpErrorStatus;
import org.snmp4j.smi.OID;
import org.snmp4j.smi.OctetString;

import java.util.Collections;
import java.util.List;

/**
 * The <code>SnmpScalarImpl</code> class implements a SNMP scalar proxy with write access.
 * @param <T>
 *   specifies the model value type of the scalar proxy.
 *
 * @author Frank Fock
 */
public class SnmpScalarImpl<T> extends SnmpScalarImplRO<T> implements SnmpCommitListener, SnmpScalar<T> {

  private SnmpProxyCommitter committer;

  /**
   * Creates a {@link org.snmp4j.model.snmp.proxy.SnmpProxyObject} for a scalar SNMP OBJECT-TYPE.
   * @param oid
   *    the OID of the scalar instance value (thus including the .0 suffix).
   * @param context
   *    the SNMP context associated with this proxy object.
   * @param value
   *    the mode value of the scalar.
   * @param valueClass
   *    the model value class.
   * @param snmpSyntax
   *    the SNMP syntax (BER syntax code) of the OBJECT-TYPE's base syntax.
   * @param committer
   *    the {@link SnmpProxyCommitter} to commit changes applied to the scalar proxy to the target SNMP entity.
   */
  public SnmpScalarImpl(OID oid, OctetString context, T value, Class<T> valueClass,
                        int snmpSyntax, SnmpProxyCommitter committer) {
    super(oid, context, value, valueClass, snmpSyntax);
    this.committer = committer;
  }

  /**
   * Change the mode value of this scalar proxy. This action will add a new {@link SnmpValuesChangeSet} to the
   * {@link SnmpProxyCommitter} if the committer is not <code>null</code>.
   * @param value
   *    the new model value for the scalar.
   */
  public void changeValue(T value) {
    if (committer != null) {
      committer.addChange(this,
          new SnmpValuesChangeSet(Collections.singletonList(
              new SnmpValueChange(getOID(), getValueClass(), value, super.getValue())), null));
    }
    super.setValue(value);
  }


  @Override
  public void commitSuccess(List<SnmpValuesChangeSet> committedChanges) {

  }

  @Override
  public void commitFailure(List<SnmpValuesChangeSet> failedChanges, SnmpErrorStatus errorStatus, int errorIndex) {

  }

  @Override
  public String toString() {
    return "SnmpScalarImpl{" +  super.toString() +
        ", committer=" + committer + "} " ;
  }
}
