/*_############################################################################
  _## 
  _##  SNMP4J-Model - SnmpSubTreeImplRO.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.proxy.impl;

import org.snmp4j.Target;
import org.snmp4j.model.mapper.VariableBindingMapper;
import org.snmp4j.model.snmp.spi.SnmpProxyReloadListener;
import org.snmp4j.model.snmp.proxy.SnmpSubTreeRO;
import org.snmp4j.model.snmp.proxy.TaskId;
import org.snmp4j.model.snmp.spi.SnmpException;
import org.snmp4j.model.snmp.spi.SnmpService;
import org.snmp4j.smi.OID;
import org.snmp4j.smi.OctetString;
import org.snmp4j.smi.VariableBinding;

import java.util.Collections;
import java.util.List;
import java.util.SortedMap;
import java.util.TreeMap;

/**
 * The <code>SnmpSubTreeImplRO</code> class implements a read-only {@link org.snmp4j.model.snmp.proxy.SnmpProxyObject}
 * for MIB sub-trees.
 * @param <T>
 *   specifies the common base value type for all model values in the sub-tree.
 *
 * @author Frank Fock
 */
public class SnmpSubTreeImplRO<T> extends SnmpProxyObjectImpl implements SnmpSubTreeRO<T> {

  private SortedMap<OID, T> values = new TreeMap<OID, T>();
  private Class<T> valueClass;

  /**
   * Creates a sub-tree proxy based on the sub-tree root {@link OID} and the common value class for the model values.
   * @param oid
   *    the root object identifier of the sub-tree.
   * @param valueClass
   *    the common value class for the model values in the sub-tree. If no common class is available, {@link Object}
   *    should be used.
   */
  public SnmpSubTreeImplRO(OID oid, OctetString context, Class<T> valueClass) {
    super(context, oid);
    this.valueClass = valueClass;
  }

  @Override
  public SortedMap<OID, T> getValues() {
    return values;
  }

  @Override
  public OID getOID() {
    return oid;
  }

  @Override
  public TaskId reload(SnmpService snmpService, Target target, VariableBindingMapper variableBindingMapper,
                       SnmpProxyReloadListener reloadListener) {
    try {
      List<? extends VariableBinding> vbs = snmpService.get(target, Collections.singletonList(getOID()));
      SortedMap<OID, T> reloaded = new TreeMap<OID, T>();
      for (VariableBinding vb : vbs) {
        reloaded.put(vb.getOid(), (T)variableBindingMapper.toObject(vb, valueClass));
      }
    } catch (SnmpException e) {
      e.printStackTrace();
    }
    return null;
  }

  @Override
  public boolean waitForTask(TaskId taskId, long timeout) {
    return false;
  }

  @Override
  public String toString() {
    return "SnmpSubTreeImplRO{" + super.toString() +
        ", valueClass=" + valueClass +
        ", values=" + values +
        "} " ;
  }
}
