/*_############################################################################
  _## 
  _##  SNMP4J-Model - SnmpValueChange.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.proxy.impl;

import org.snmp4j.model.snmp.proxy.ChangeSet;
import org.snmp4j.model.snmp.spi.SnmpErrorStatus;
import org.snmp4j.smi.OID;
import org.snmp4j.smi.VariableBinding;

import java.util.Collections;
import java.util.List;

/**
 * The <code>SnmpValueChange</code> class represents a single value change with old a new value
 * supporting undo and redo operations.
 *
 * @author Frank Fock
 */
public class SnmpValueChange extends VariableBinding implements ChangeSet<SnmpValueChange> {

  private Object newValue;
  private Object oldValue;
  private Class valueClass;

  private ChangeStatus changeStatus;
  private SnmpErrorStatus errorStatus;

  /**
   * Creates a value change with SNMP instance OID, model value class, old, and new value.
   * @param oid
   *   the instance OID of the target SNMP object instance (i.e., for scalars including the <code>.0</code> suffix).
   * @param valueClass
   *   the value class of the model values.
   * @param newValue
   *   the new model value.
   * @param oldValue
   *   the old model value (or <code>null</code> if it did not exist).
   */
  public SnmpValueChange(OID oid, Class valueClass, Object newValue, Object oldValue) {
    super(oid);
    this.valueClass = valueClass;
    this.newValue = newValue;
    this.oldValue = oldValue;
    this.changeStatus = ChangeStatus.PENDING;
  }

  /**
   * Gets the class of the model values.
   * @return
   *   the class of the model values.
   */
  public Class getValueClass() {
    return valueClass;
  }

  /**
   * Gets the status of the change as defined by {@link org.snmp4j.model.snmp.proxy.ChangeSet.ChangeStatus}.
   * @return
   *    the status of the change.
   */
  public ChangeStatus getChangeStatus() {
    return changeStatus;
  }

  /**
   * Gets the new value.
   * @return
   *    the new value.
   */
  public Object getNewValue() {
    return newValue;
  }

  /**
   * Sets the new value.
   * @param newValue
   *   the new value (the value class must be compatible with the {@link #getValueClass()} otherwise the
   *   mapping to the SNMP value and vice versa will fail (later).
   */
  public void setNewValue(Object newValue) {
    this.newValue = newValue;
  }

  /**
   * Gets the old value.
   * @return
   *    the old value.
   */
  public Object getOldValue() {
    return oldValue;
  }

  /**
   * Sets the old value.
   * @param oldValue
   *   the old value (the value class must be compatible with the {@link #getValueClass()} otherwise the
   *   mapping to the SNMP value and vice versa will fail (later).
   */
  public void setOldValue(Object oldValue) {
    this.oldValue = oldValue;
  }

  @Override
  public List<SnmpValueChange> getChanges() {
    return Collections.singletonList(this);
  }

  @Override
  public ChangeStatus getStatus() {
    return changeStatus;
  }

  @Override
  public void setStatus(ChangeStatus status) {
    this.changeStatus = changeStatus;
  }

  /**
   * Gets the SNMP error status for this change.
   * @return
   *    the SNMP error status as defined by {@link SnmpErrorStatus}.
   */
  public SnmpErrorStatus getErrorStatus() {
    return errorStatus;
  }

  /**
   * Sets the SNMP error status for this change.
   * @param errorStatus
   *    the new SNMP error status as defined by {@link SnmpErrorStatus}.
   */
  public void setErrorStatus(SnmpErrorStatus errorStatus) {
    this.errorStatus = errorStatus;
  }

  @Override
  public String toString() {
    return "SnmpValueChange{" +
        "oid="+ getOid() +
        ", valueClass=" + valueClass +
        ", newValue=" + newValue +
        ", oldValue=" + oldValue +
        ", smiSyntax=" + getSyntax() +
        ", changeStatus=" + changeStatus +
        ", errorStatus=" + errorStatus +
        '}';
  }
}
