/*_############################################################################
  _## 
  _##  SNMP4J-Model - SnmpValuesChangeSet.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.proxy.impl;

import org.snmp4j.model.snmp.proxy.ChangeSet;
import org.snmp4j.model.snmp.spi.SnmpErrorStatus;
import org.snmp4j.smi.VariableBinding;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * The SnmpValuesChangeSet class implements a set of changes that must be
 * applied atomically.
 *
 * @author Frank Fock
 */
public class SnmpValuesChangeSet implements ChangeSet<SnmpValueChange> {

  private List<SnmpValueChange> changes;
  private ChangeStatus status;
  private Object userReference;

  /**
   * Creates an empty change set.
   */
  public SnmpValuesChangeSet() {
    this.changes = new ArrayList<SnmpValueChange>();
  }

  /**
   * Creates a change set based on a provided list of changes and with an associated user object.
   * @param changes
   *   the changes to include in the change set. The provided collection is not be used internally.
   * @param userReference
   *   a arbitrary user defined object to be associated with this change set.
   */
  public SnmpValuesChangeSet(Collection<SnmpValueChange> changes, Object userReference) {
    this();
    this.changes.addAll(changes);
    this.userReference = userReference;
  }

  /**
   * Adds a change to the change set.
   * @param change
   *    the new change to add.
   */
  public void addChange(SnmpValueChange change) {
    this.changes.add(change);
  }

  @Override
  public List<SnmpValueChange> getChanges() {
    return changes;
  }

  /**
   * Gets all the new values as new a list.
   * @return
   *    a list of the new values in the change set.
   */
  @SuppressWarnings("unchecked")
  public List getNewValues() {
    List newValues = new ArrayList();
    for (SnmpValueChange change : changes) {
      newValues.add(change.getNewValue());
    }
    return newValues;
  }

  /**
   * Checks if all variables are not <code>null/code>.
   * @return
   *   <code>true</code> if all SNMP variables are set (not null).
   */
  public boolean isVariableSetOk() {
    for (SnmpValueChange change : changes) {
      if (change.getVariable() == null) {
        return false;
      }
    }
    return true;
  }

  /**
   * Gets the variable bindings that are formed by this change set.
   * @return
   *    a list of SNMP variable bindings.
   */
  public List<? extends VariableBinding> getVariableBindings() {
    return changes;
  }

  @Override
  public ChangeStatus getStatus() {
    return status;
  }

  @Override
  public void setStatus(ChangeStatus status) {
    this.status = status;
  }

  @Override
  public SnmpErrorStatus getErrorStatus() {
    for (SnmpValueChange change : changes) {
      if (change.getErrorStatus() != null) {
        return change.getErrorStatus();
      }
    }
    return null;
  }

  /**
   * Gets the user reference object associated with this change.
   * @return
   *    the user reference object.
   */
  public Object getUserReference() {
    return userReference;
  }

  @Override
  public String toString() {
    return "SnmpValuesChangeSet{" +
        "changes=" + changes +
        ", status=" + status +
        ", userReference=" + userReference +
        '}';
  }
}
