/*_############################################################################
  _## 
  _##  SNMP4J-Model - SnmpService.java  
  _## 
  _##  Copyright (C) 2013-2015  Frank Fock (SNMP4J.org)
  _##  
  _##  This program is free software; you can redistribute it and/or modify
  _##  it under the terms of the GNU General Public License version 2 as 
  _##  published by the Free Software Foundation.
  _##
  _##  This program is distributed in the hope that it will be useful,
  _##  but WITHOUT ANY WARRANTY; without even the implied warranty of
  _##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  _##  GNU General Public License for more details.
  _##
  _##  You should have received a copy of the GNU General Public License
  _##  along with this program; if not, write to the Free Software
  _##  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
  _##  MA  02110-1301  USA
  _##  
  _##########################################################################*/
package org.snmp4j.model.snmp.spi;

import org.snmp4j.Target;
import org.snmp4j.model.snmp.proxy.SnmpCommitListener;
import org.snmp4j.model.snmp.proxy.SnmpTableUpdateListener;
import org.snmp4j.model.snmp.proxy.impl.SnmpValuesChangeSet;
import org.snmp4j.model.transaction.SnmpTransaction;
import org.snmp4j.model.transaction.TransactionManager;
import org.snmp4j.smi.OID;
import org.snmp4j.smi.VariableBinding;

import java.io.IOException;
import java.util.List;

/**
 * The <code>SnmpService</code> provides SNMP services to SNMP command generator applications.
 *
 * @author Frank Fock
 */
public interface SnmpService extends TransactionManager {

  /**
   * Initializes the transports associated with this service. This will open the necessary network ports.
   * @throws IOException
   *    if an IO error occurred during initialization.
   */
  void initTransports() throws IOException;

  /**
   * Gets the SNMP values for the specified OIDs.
   * @param target
   *   the SNMP target to send the SNMP PDUs to.
   * @param oidList
   *   the instance object identifiers to GET values for.
   * @return
   *   a list of variable bindings.
   * @throws SnmpException
   *   if an SNMP error or timeout occurred.
   */
  List<? extends VariableBinding> get(Target target, List<OID> oidList) throws SnmpException;

  /**
   * Gets the next following SNMP values for the specified OIDs.
   * @param target
   *   the SNMP target to send the SNMP PDUs to.
   * @param oidList
   *   the instance object identifiers to GET-NEXT values for.
   * @return
   *   a list of variable bindings.
   * @throws SnmpException
   *   if an SNMP error or timeout occurred.
   */
  List<? extends VariableBinding> getNext(Target target, List<OID> oidList) throws SnmpException;

  /**
   * Gets the SNMP values for the specified sub-tree root OID.
   * @param target
   *   the SNMP target to send the SNMP PDUs to.
   * @param oid
   *   the root OID of the sub-tree to get values for (using GETBULK or GETNEXT).
   * @return
   *   a list of variable bindings.
   * @throws SnmpException
   *   if an SNMP error or timeout occurred.
   */
  List<? extends VariableBinding> getSubTree(Target target, OID oid) throws SnmpException;

  /**
   * Gets the rows of a SNMP table synchronously.
   * @param target
   *   the SNMP target to send the SNMP PDUs to.
   * @param columnOIDs
   *   the OIDs of the columns to retrieve (columns might be combined from several tables although the indexes
   *   of the included tables should then contain a common prefix).
   * @param minIndex
   *   the minimum index to return (or <code>null</code> for no limit).
   * @param maxIndex
   *   the maximum index to return (or <code>null</code> for no limit).
   * @return
   *   a list of {@link SnmpRow} instances.
   * @throws SnmpException
   *   if an SNMP error or timeout occurred.
   */
  List<? extends SnmpRow> getTable(Target target, List<OID> columnOIDs, OID minIndex, OID maxIndex)
      throws SnmpException;

  /**
   * Gets the rows of a SNMP table asynchronously.
   * @param target
   *   the SNMP target to send the SNMP PDUs to.
   * @param columnOIDs
   *   the OIDs of the columns to retrieve (columns might be combined from several tables although the indexes
   *   of the included tables should then contain a common prefix).
   * @param minIndex
   *   the minimum index to return (or <code>null</code> for no limit).
   * @param maxIndex
   *   the maximum index to return (or <code>null</code> for no limit).
   * @param callback
   *   the callback that receives the retrieved rows.
   * @param userObject
   *   an optinal user object to be associated with this retrieval.
   * @throws SnmpException
   *   if an SNMP error or timeout occurred.
   */
  void getTable(Target target, List<OID> columnOIDs, OID minIndex, OID maxIndex,
                SnmpRowListener callback, Object userObject) throws SnmpException;

  /**
   * Updates (SETs) a set of changes with an atomic SET operation asynchronously.
   * @param snmpTransaction
   *   the transaction that controls this update.
   * @param target
   *   the SNMP target.
   * @param changeSet
   *   the change set to commit to the agent identified by <code>target</code>.
   * @param snmpCommitListener
   *   a listener for commit events.
   */
  void update(SnmpTransaction snmpTransaction, Target target, SnmpValuesChangeSet changeSet,
              SnmpCommitListener snmpCommitListener);
}
